"""
Module for gathering data that will be combined to calculate transfer functions

There are two scenarios considered here. The first is the simplest, which is
quick processing outside the project environment. In this case data gathering
is not complicated. This workflow does not involve a data selector, meaining
only a single step is required.

- QuickGather to put together the out_data, in_data and cross_data

When inside the project environment, regardless of whether it is single site or
multi site processing, the workflow follows:

- Selector to select shared windows across all sites for a sampling frequency
- Gather to gather the combined evaluation frequency data

.. warning::

    There may be some confusion in the code with many references to spectra data
    and evaluation frequency data. Evaluation frequency data, referred to below
    as eval_data is actually an instance of Spectra data. However, it is named
    differently to highlight the fact that it is not the complete spectra data,
    but is actually spectra data at a reduced set of frequencies corresponding
    to the evaluation frequncies.

    Within a project, there are separate folders for users who want to save both
    the full spectra data with all the frequencies as well as the evaluation
    frequency spectra data with the smaller subset of frequencies. Only the
    evaluation frequency data is required to calculate the transfer function,
    but the full spectral data might be useful for visualisation and analysis
    reasons.
"""
from loguru import logger
from typing import List, Dict, Optional, Tuple
from pathlib import Path
import numpy as np
import pandas as pd

from resistics.common import ResisticsProcess, ResisticsData
from resistics.common import WriteableMetadata, History
from resistics.project import Project, Site
from resistics.decimate import DecimationParameters
from resistics.spectra import SpectraLevelMetadata, SpectraMetadata, SpectraData
from resistics.spectra import SpectraDataReader
from resistics.transfunc import TransferFunction


def get_site_evals_metadata(
    config_name: str, proj: Project, site_name: str, fs: float
) -> Dict[str, SpectraMetadata]:
    """
    Get spectra metadata for a given site and sampling frequency

    Parameters
    ----------
    config_name : str
        The configuration name to get the right data
    proj : Project
        The project instance to get the measurements
    site_name : str
        The name of the site for which to gather the SpectraMetadata
    fs : float
        The original recording sampling frequency

    Returns
    -------
    Dict[str, SpectraMetadata]
        Dictionary of measurement name to SpectraMetadata
    """
    from resistics.project import get_meas_evals_path

    site = proj[site_name]
    measurements = site.get_measurements(fs=fs)
    meas_metadata = {}
    for meas_name in measurements:
        meas = site[meas_name]
        evals_path = get_meas_evals_path(
            proj.dir_path, site.name, meas.name, config_name
        )
        try:
            metadata = SpectraDataReader().run(evals_path, metadata_only=True)
        except Exception:
            logger.error(f"No evals data found in path {evals_path}")
            continue
        logger.info(f"Found evals data for {site.name}, {meas.name}, {config_name}")
        meas_metadata[meas.name] = metadata
    return meas_metadata


def get_site_level_wins(
    meas_metadata: Dict[str, SpectraMetadata], level: int
) -> pd.Series:
    """
    Get site windows for a decimation level given a sampling frequency

    Parameters
    ----------
    meas_metadata : Dict[str, SpectraMetadata]
        The measurement spectra metadata for a site
    level : int
        The decimation level

    Returns
    -------
    pd.Series
        A series with an index of global windows for the site and values the
        measurements which have that global window. This is for a single
        decimation level

    See Also
    --------
    get_site_wins : Get windows for all decimation levels

    Examples
    --------
    An example getting the site windows for decimation level 0 when there are
    three measurements in the site.

    >>> from resistics.testing import spectra_metadata_multilevel
    >>> from resistics.gather import get_site_level_wins
    >>> meas_metadata = {}
    >>> meas_metadata["meas1"] = spectra_metadata_multilevel(n_wins=[3, 2, 2], index_offset=[3, 2, 1])
    >>> meas_metadata["meas2"] = spectra_metadata_multilevel(n_wins=[4, 3, 2], index_offset=[28, 25, 22])
    >>> meas_metadata["meas3"] = spectra_metadata_multilevel(n_wins=[2, 2, 1], index_offset=[108, 104, 102])
    >>> get_site_level_wins(meas_metadata, 0)
    3      meas1
    4      meas1
    5      meas1
    28     meas2
    29     meas2
    30     meas2
    31     meas2
    108    meas3
    109    meas3
    dtype: object
    >>> get_site_level_wins(meas_metadata, 1)
    2      meas1
    3      meas1
    25     meas2
    26     meas2
    27     meas2
    104    meas3
    105    meas3
    dtype: object
    >>> get_site_level_wins(meas_metadata, 2)
    1      meas1
    2      meas1
    22     meas2
    23     meas2
    102    meas3
    dtype: object
    """
    level_wins = pd.Series(dtype=str)
    for meas_name, metadata in meas_metadata.items():
        if level >= metadata.n_levels:
            continue
        level_metadata = metadata.levels_metadata[level]
        first_global = level_metadata.index_offset
        last_gobal = level_metadata.index_offset + level_metadata.n_wins - 1
        index = np.arange(first_global, last_gobal + 1)
        level_wins = level_wins.append(pd.Series(data=meas_name, index=index))
    level_wins = level_wins[~level_wins.index.duplicated(keep="first")]
    return level_wins.sort_index()


def get_site_wins(
    config_name: str, proj: Project, site_name: str, fs: float
) -> Dict[int, pd.Series]:
    """
    Get site windows for all levels given a sampling frequency

    Parameters
    ----------
    config_name : str
        The configuration name to get the right data
    proj : Project
        The project instance to get the measurements
    site_name : str
        The site name
    fs : float
        The recording sampling frequency

    Returns
    -------
    Dict[int, pd.Series]
        Dictionary of integer to levels, with one entry for each decimation
        level

    Raises
    ------
    ValueError
        If no matching spectra metadata is found
    """
    logger.debug(f"Getting windows for site {site_name}")
    meas_metadata = get_site_evals_metadata(config_name, proj, site_name, fs)
    if len(meas_metadata) == 0:
        raise ValueError(f"No measurements for site {site_name}, sample frequency {fs}")
    n_levels = max([x.n_levels for x in meas_metadata.values()])
    logger.debug(f"Levels for site {site_name}, sample frequency {fs} = {n_levels}")
    tables = {}
    for ilevel in range(n_levels):
        tables[ilevel] = get_site_level_wins(meas_metadata, ilevel)
    return tables


class Selection(ResisticsData):
    """
    Selections are output by the Selector. They hold information about the data
    that should be gathered for the regression.
    """

    def __init__(
        self,
        sites: List[Site],
        dec_params: DecimationParameters,
        tables: Dict[int, pd.DataFrame],
    ):
        """
        Initialise the selection

        Parameters
        ----------
        sites : List[Site]
            The list of Sites that were included in the Selector
        dec_params : DecimationParameters
            The decimation parameters
        tables : Dict[int, pd.DataFrame]
            The window tables
        """
        self.sites = sites
        self.dec_params = dec_params
        self.tables = tables
        self.n_levels = len(tables)

    def get_n_evals(self) -> int:
        """
        Get the total number of evaluation frequnecies

        Returns
        -------
        int
            The total number of evaluation frequencies that can be calculated
        """
        return self.n_levels * self.dec_params.per_level

    def get_n_wins(self, level: int, eval_idx: int) -> int:
        """
        Get the number of windows for an evaluation frequency

        Parameters
        ----------
        level : int
            The decimation level
        eval_idx : int
            The evaluation frequency index in the decimation level

        Returns
        -------
        int
            The number of windows

        Raises
        ------
        ValueError
            If the level is greater than the maximum level available
        """
        if level >= self.n_levels:
            raise ValueError(f"Level {level} >= n_levels {self.n_levels}")
        level_table = self.tables[level]
        eval_series = level_table[eval_idx]
        return eval_series[eval_series].count()

    def get_measurements(self, site: Site) -> List[str]:
        """
        Get the measurement names to read from a Site

        Parameters
        ----------
        site : Site
            The site for which to get the measurements

        Returns
        -------
        List[str]
            The measurements to read from
        """
        measurements = set()
        for level_table in self.tables.values():
            level_set = set(level_table[site.name].unique())
            measurements = measurements.union(level_set)
        return sorted(list(measurements))

    def get_eval_freqs(self) -> List[float]:
        """
        Get the evaluation frequencies

        Returns
        -------
        List[float]
            The evaluation frequencies as a flat list of floats
        """
        eval_freqs = []
        for ilevel in range(self.n_levels):
            eval_freqs = eval_freqs + self.dec_params.get_eval_freqs(ilevel)
        return eval_freqs

    def get_eval_wins(self, level: int, eval_idx: int) -> pd.DataFrame:
        """
        Limit the level windows to the evaluation frequency

        Parameters
        ----------
        level : int
            The decimation level
        eval_idx : int
            The evalution frequency index in the decimation level

        Returns
        -------
        pd.DataFrame
            pandas DataFrame of the windows and the measurements from each site
            the window can be read from
        """
        cols = [site.name for site in self.sites]
        eval_wins = self.tables[level]
        eval_wins = eval_wins[eval_wins[eval_idx]]
        return eval_wins[cols]


class Selector(ResisticsProcess):
    """
    The Selector takes Sites and tries to find shared windows across them. A
    project instance is required for the Selector to be able to find shared
    windows.

    The Selector should be used for remote reference and intersite processing
    and single site processing when masks are involved.
    """

    def run(
        self,
        config_name: str,
        proj: Project,
        site_names: List[str],
        dec_params: DecimationParameters,
        masks: Optional[Dict[str, str]] = None,
    ) -> Selection:
        """
        Run the selector

        If a site repeats, the selector only considers it once. This might be
        the case when performing intersite or other cross site style processing.

        Parameters
        ----------
        config_name : str
            The configuration name
        proj : Project
            The project instance
        site_names : List[str]
            The names of the sites to get data from
        dec_params : DecimationParameters
            The decimation parameters with number of levels etc.
        masks : Optional[Dict[str, str]], optional
            Any masks to add, by default None

        Returns
        -------
        Selection
            The Selection information defining the measurements and windows to
            read for each site
        """
        # get unique sites
        site_names = sorted(list(set(site_names)))
        fs = dec_params.fs
        sites_wins = {
            site_name: get_site_wins(config_name, proj, site_name, fs)
            for site_name in site_names
        }
        # get the higest decimation level that all sites have
        n_levels = min([len(x) for x in sites_wins.values()])
        logger.info(f"Finding shared windows across {', '.join(sites_wins.keys())}")
        logger.info(f"Max. level across sites = {n_levels - 1}, num. levels {n_levels}")
        tables: Dict[int, pd.DataFrame] = {}
        for ilevel in range(n_levels):
            logger.info(f"Finding shared windows for decimation level {ilevel}")
            data = {x: y[ilevel] for x, y in sites_wins.items()}
            table = pd.DataFrame(data=data).dropna()
            table = self._get_evals(dec_params, table, ilevel)
            if masks is not None:
                table = self._apply_masks(table, masks)
            tables[ilevel] = table
        sites = [proj[site_name] for site_name in site_names]
        return Selection(sites, dec_params, tables)

    def _get_evals(
        self, dec_params: DecimationParameters, table: pd.DataFrame, level: int
    ) -> pd.DataFrame:
        """
        Add a column for each evaluation frequency

        Parameters
        ----------
        dec_params : DecimationParameters
            The decimation parameters
        table : pd.DataFrame
            The window table with measurements from each site
        level : int
            The decimation level

        Returns
        -------
        pd.DataFrame
            pandas DataFrame with boolean column for each evaluation frequency
        """
        for ifreq in range(dec_params.per_level):
            table[ifreq] = True
        return table

    def _apply_masks(self, table: pd.DataFrame, masks: Dict[str, str]) -> pd.DataFrame:
        """Set some windows False based on masks"""
        return table


class SiteCombinedMetadata(WriteableMetadata):
    """
    Metadata for combined data

    Combined metadata stores metadata for measurements that are combined from
    a single site.
    """

    site_name: str
    """The name of the site"""
    fs: float
    """Recording sampling frequency"""
    system: str = ""
    """The system used for recording"""
    serial: str = ""
    """Serial number of the system"""
    wgs84_latitude: float = -999.0
    """Latitude in WGS84"""
    wgs84_longitude: float = -999.0
    """Longitude in WGS84"""
    easting: float = -999.0
    """The easting of the site in local cartersian coordinates"""
    northing: float = -999.0
    """The northing of the site in local cartersian coordinates"""
    elevation: float = -999.0
    """The elevation of the site"""
    measurements: Optional[List[str]] = None
    """List of measurement names that were included in the combined data"""
    chans: List[str]
    """List of channels, these are common amongst all the measurements"""
    n_evals: int
    """The number of evaluation frequencies"""
    eval_freqs: List[float]
    """The evaluation frequencies"""
    histories: Dict[str, History]
    """Dictionary mapping measurement name to measurement processing history"""


class SiteCombinedData(ResisticsData):
    """
    Combined data is data that is combined from a single site for the purposes
    of regression.

    All of the data that is combined should have the same sampling frequency,
    same evaluation frequencies and some shared channels.

    Data is stored in the data attribute of the class. This is a dictionary
    mapping evaluation frequency index to data for the evaluation frequency
    from all windows in the site. The shape of data for a single evaluation
    frequency is:

    n_wins x n_chans

    The data is complex valued.
    """

    def __init__(self, metadata: SiteCombinedMetadata, data: Dict[int, np.ndarray]):
        """
        Initialise the CombinedData

        Parameters
        ----------
        metadata : CombinedMetadata
            The combined metadata
        data : Dict[int, np.ndarray]
            The data with index the evaluation frequency index and value the
            combined data from a site for the evaluation frequency
        """
        self.metadata = metadata
        self.data = data


class GatheredData(ResisticsData):
    """
    Class to hold data to be used in by Regression preparers

    Gathered data has an out_data, in_data and cross_data. The important thing
    here is that the data is all aligned with regards to windows
    """

    def __init__(
        self,
        out_data: SiteCombinedData,
        in_data: SiteCombinedData,
        cross_data: SiteCombinedData,
    ) -> None:
        self.out_data = out_data
        self.in_data = in_data
        self.cross_data = cross_data


class ProjectGather(ResisticsProcess):
    """
    Gather aligned data from a single or multiple sites in the project

    Aligned data means that the same index of data across multiple sites points
    to data covering the same global window (i.e. the same time window). This
    is essential for calculating intersite or remote reference transfer
    functions.
    """

    def run(
        self,
        config_name: str,
        proj: Project,
        selection: Selection,
        tf: TransferFunction,
        out_name: str,
        in_name: Optional[str] = None,
        cross_name: Optional[str] = None,
    ) -> GatheredData:
        """
        Gather data for input into the regression preparer

        Parameters
        ----------
        config_name : str
            The config name for getting the correct evals data
        proj : Project
            The project instance
        selection : Selection
            The selection
        tf : TransferFunction
            The transfer function
        out_name : str
            The name of the output site
        in_name : Optional[str], optional
            The name of the input site, by default None
        cross_name : Optional[str], optional
            The name of the cross site, by default None

        Returns
        -------
        GatheredData
            The data gathered for the regression preparer
        """
        if in_name is None:
            in_name = out_name
        if cross_name is None:
            cross_name = in_name

        # collect the data
        out_data = self._get_site_data(
            config_name, proj, selection, out_name, tf.out_chans
        )
        in_data = self._get_site_data(
            config_name, proj, selection, in_name, tf.in_chans
        )
        cross_data = self._get_site_data(
            config_name, proj, selection, cross_name, tf.cross_chans
        )
        return GatheredData(out_data, in_data, cross_data)

    def _get_site_data(
        self,
        config_name: str,
        proj: Project,
        selection: Selection,
        site_name: str,
        chans: List[str],
    ) -> SiteCombinedData:
        """
        Collect the evals data for the site

        Parameters
        ----------
        config_name : str
            The configuration name to fetch the correct data
        proj : Project
            The project instance
        selection : Selection
            The window selection
        site_name : str
            The site name
        chans : List[str]
            The channels to get for the site

        Returns
        -------
        SiteCombinedData
            A combined data instance
        """
        from resistics.project import get_meas_evals_path

        site = proj[site_name]
        measurements = selection.get_measurements(site)
        data = self._get_empty_data(selection, chans)
        histories = {}
        metadata = None
        for meas in measurements:
            evals_path = get_meas_evals_path(
                proj.dir_path, site.name, meas, config_name
            )
            eval_data = SpectraDataReader().run(evals_path)
            self._populate_data(selection, site, meas, eval_data, chans, data)
            histories[meas] = eval_data.metadata.history
            if metadata is None:
                metadata = eval_data.metadata
        combined_metadata = SiteCombinedMetadata(
            site_name=site.name,
            fs=selection.dec_params.fs,
            system=metadata.system,
            serial=metadata.serial,
            wgs84_latitude=metadata.wgs84_latitude,
            wgs84_longitude=metadata.wgs84_longitude,
            easting=metadata.easting,
            northing=metadata.northing,
            elevation=metadata.elevation,
            measurements=measurements,
            chans=chans,
            n_evals=len(data),
            eval_freqs=selection.get_eval_freqs(),
            histories=histories,
        )
        return SiteCombinedData(combined_metadata, data)

    def _get_empty_data(
        self, selection: Selection, chans: List[str]
    ) -> Dict[int, np.ndarray]:
        """
        Get dictionary of empty arrays to put the data in

        This is a dictionary with n_evals entries
        Each evaluation frequency entry is of size:

        n_wins * n_chans * 1

        Parameters
        ----------
        selection : Selection
            The window selection information
        chans : List[str]
            The channels to define the shape of the array

        Returns
        -------
        Dict[int, np.ndarray]
            Dictionary mapping evaluation frequency index to the combined
            evaluation frequency data
        """
        per_level = selection.dec_params.per_level
        empty_data = {}
        for ilevel in range(selection.n_levels):
            for ifreq in range(per_level):
                n_wins = selection.get_n_wins(ilevel, ifreq)
                key = per_level * ilevel + ifreq
                empty_data[key] = np.empty(
                    shape=(n_wins, len(chans)), dtype=np.complex128
                )
        return empty_data

    def _populate_data(
        self,
        selection: Selection,
        site: Site,
        meas: str,
        eval_data: SpectraData,
        chans: List[str],
        data: Dict[int, np.ndarray],
    ) -> Dict[int, np.ndarray]:
        """
        Populate a measurement's evaluation frequency data into combined data

        There is some complexity here regarding mapping of the right windows to
        the correct indices for the site.

        Parameters
        ----------
        selection : Selection
            The selection data
        site : Site
            The site
        meas : str
            The name of the measurement
        eval_data : SpectraData
            The evaluation frequency data for the measurement
        chans : List[str]
            The channels to get from the data
        data : Dict[int, np.ndarray]
            The data dictionary which will be populated

        Returns
        -------
        Dict[int, np.ndarray]
            The data dictionary with data for the measurement populated
        """
        per_level = selection.dec_params.per_level
        for ilevel in range(selection.n_levels):
            if ilevel >= eval_data.metadata.n_levels:
                logger.debug(f"Measurement {meas} has no level {ilevel}")
                break
            level_data = eval_data.get_chans(ilevel, chans)
            for ifreq in range(per_level):
                key = per_level * ilevel + ifreq
                eval_wins = selection.get_eval_wins(ilevel, ifreq)
                eval_data_indices, combined_indices = self._get_indices(
                    eval_wins, site, meas, eval_data.metadata.levels_metadata[ilevel]
                )
                data[key][combined_indices] = level_data[eval_data_indices, ..., ifreq]
        return data

    def _get_indices(
        self,
        eval_wins: pd.DataFrame,
        site: Site,
        meas_name: str,
        level_metadata: SpectraLevelMetadata,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get two arrays to help align windows

        - eval_data_indices are the the global indices relative to measurement
        - combined_indices are indices relative to the combined data

        Breaking this down even more, suppose there are three measurements in
        a site at a sampling frequency of 10 Hz. Together, these measurements
        will span a range of global indices with gaps in between, for example:

        - Measurement 1: 0, 1, 2, 3, 4
        - Measurement 2: 20, 21, 22, 23, 24
        - Measurement 3: 30, 31, 32, 33, 34

        Relative to their own start times, the indices will be:

        - Measurement 1: 0, 1, 2, 3, 4
        - Measurement 2: 0, 1, 2, 3, 4
        - Measurement 3: 0, 1, 2, 3, 4

        Finally, when combining data for a site, these need to be reindexed
        relative to the number of contributing windows in the site.

        - Measurement 1: 0, 1, 2, 3, 4
        - Measurement 2: 5, 6, 7, 8, 9
        - Measurement 3: 10, 11, 12, 13, 14

        This method returns the mapping from the index relative to the
        measurement to the index relative to the site. For Measurement 2, this
        would be:

        0, 1, 2, 3, 4 -> 5, 6, 7, 8, 9

        Parameters
        ----------
        eval_wins : pd.DataFrame
            The global windows for the evaluation frequencies
        site : Site
            The site instance
        meas_name : str
            The name of the measurement
        level_metadata : SpectraLevelMetadata
            The spectra level metadata

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Indices relative to the spectra data and indices relative to the
            site combined data
        """
        eval_wins["combined_index"] = np.arange(len(eval_wins))
        eval_meas_wins = eval_wins[eval_wins[site.name] == meas_name]
        eval_data_indices = eval_meas_wins.index.values - level_metadata.index_offset
        combined_indices = eval_meas_wins["combined_index"].values
        return eval_data_indices, combined_indices


class QuickGather(ResisticsProcess):
    """
    Processor to gather data outside of a resistics environment

    This is intended for use when quickly calculating out a transfer function
    for a single measurement and only a single spectra data instance is accepted
    as input.

    Remote reference or intersite processing is not possible using QuickGather

    See Also
    --------
    ProjectGather : For more advanced gathering of data in a project
    """

    def run(
        self,
        dir_path: Path,
        dec_params: DecimationParameters,
        tf: TransferFunction,
        eval_data: SpectraData,
    ) -> GatheredData:
        """
        Generate the GatheredData object for input into regression preparation

        The input is a single spectra data instance and is used to populate the
        in_data, out_data and cross_data.

        Parameters
        ----------
        dir_path : Path
            The directory path to the measurement
        dec_params : DecimationParameters
            The decimation parameters
        tf : TransferFunction
            The transfer function
        eval_data : SpectraData
            The spectra data at the evaluation frequencies

        Returns
        -------
        GatheredData
            GatheredData for regression preparer
        """
        metadata = eval_data.metadata
        out_data = {}
        in_data = {}
        cross_data = {}
        eval_freqs = []
        logger.info("Quick gathering data for regression prepartion")
        for ilevel in range(metadata.n_levels):
            level_metadata = metadata.levels_metadata[ilevel]
            eval_freqs = eval_freqs + level_metadata.freqs
            out_level = eval_data.get_chans(ilevel, tf.out_chans)
            in_level = eval_data.get_chans(ilevel, tf.in_chans)
            cross_level = eval_data.get_chans(ilevel, tf.cross_chans)
            for ifreq in range(level_metadata.n_freqs):
                key = dec_params.per_level * ilevel + ifreq
                out_data[key] = out_level[..., ifreq]
                in_data[key] = in_level[..., ifreq]
                cross_data[key] = cross_level[..., ifreq]
        # make combined data
        fs = dec_params.fs
        out_combined = self._get_combined_data(
            dir_path.name, fs, tf.out_chans, eval_freqs, metadata, out_data
        )
        in_combined = self._get_combined_data(
            dir_path.name, fs, tf.in_chans, eval_freqs, metadata, in_data
        )
        cross_combined = self._get_combined_data(
            dir_path.name, fs, tf.cross_chans, eval_freqs, metadata, cross_data
        )
        return GatheredData(
            out_data=out_combined, in_data=in_combined, cross_data=cross_combined
        )

    def _get_combined_data(
        self,
        meas: str,
        fs: float,
        chans: List[str],
        eval_freqs: List[float],
        metadata: SpectraMetadata,
        data: Dict[int, np.ndarray],
    ) -> SiteCombinedData:
        """Get the combined metadata"""
        combined_metadata = SiteCombinedMetadata(
            site_name=meas,
            fs=fs,
            system=metadata.system,
            serial=metadata.serial,
            wgs84_latitude=metadata.wgs84_latitude,
            wgs84_longitude=metadata.wgs84_longitude,
            easting=metadata.easting,
            northing=metadata.northing,
            elevation=metadata.elevation,
            measurements=[meas],
            chans=chans,
            n_evals=len(eval_freqs),
            eval_freqs=eval_freqs,
            histories={meas: metadata.history},
        )
        return SiteCombinedData(combined_metadata, data)



"""
The regression module provides functions and classes for the following:

- Preparing gathered data for regression
- Performing the linear regression

Resistics has built in solvers that use scikit learn models, namely

- Ordinary least squares
- RANSAC
- TheilSen

These will perform well in many scenarios. However, the functionality available
in resistics makes it possible to use custom solvers if required.
"""
from loguru import logger
from typing import List, Dict, Tuple, Union, Optional
from tqdm import tqdm
import numpy as np
from sklearn.base import BaseEstimator

from resistics.common import Metadata, WriteableMetadata, History
from resistics.common import ResisticsData, ResisticsProcess
from resistics.transfunc import Component, get_component_key
from resistics.transfunc import TransferFunction
from resistics.spectra import SpectraMetadata, SpectraData
from resistics.gather import SiteCombinedMetadata, GatheredData


class RegressionInputMetadata(Metadata):
    """Metadata for regression input data, mainly to track processing history"""

    contributors: Dict[str, Union[SiteCombinedMetadata, SpectraMetadata]]
    """Details about the data contributing to the regression input data"""
    history: History = History()
    """The processing history"""


class RegressionInputData(ResisticsData):
    """
    Class to hold data that will be input into a solver

    The purpose of regression input data is to provision for many different
    solvers and user written solvers.

    The regression input data has the following key attributes:

    - freqs
    - obs
    - preds

    The freqs attribute is a 1-D array of evaluation frequencies.

    The obs attribute is a dictionary of dictionaries. The parent dictionary has
    a key of the evaluation frequency index. The secondary dictionary has key of
    output channel. The values in the secondary dictionary are the observations
    for that output channel and have 1-D size:

    (n_wins x n_cross_chans x 2).

    The factor of 2 is because the real and complex parts of each equation are
    separated into two equations to allow use of solvers that work exclusively
    on real data.

    The preds attribute is a single level dictionary with key of evaluation
    frequency index and value of the predictors for the evaluation frequency.
    The predictors have 2-D shape:

    (n_wins x n_cross_chans x 2)  x (n_input_channels x 2).

    The number of windows is multiplied by 2 for the same reason as the
    observations. The doubling of the input channels is because one is the
    predictor for the real part of that transfer function component and one is
    the predictor for the complex part of the transfer function component.

    Considering the impedance tensor as an example with:

    - output channels Ex, Ey
    - input channels Hx, Hy
    - cross channels Hx, Hy

    The below shows the arrays for the 0 index evaluation frequency:

    Observations

    - Ex: [w1_crossHx_RE, w1_crossHx_IM, w1_crossHy_RE, w1_crossHy_IM]
    - Ey: [w1_crossHx_RE, w1_crossHx_IM, w1_crossHy_RE, w1_crossHy_IM]

    Predictors Ex

    - w1_crossHx_RE:    Zxx_RE  Zxx_IM  Zxy_RE  Zxy_IM
    - w1_crossHx_IM:    Zxx_RE  Zxx_IM  Zxy_RE  Zxy_IM
    - w1_crossHy_RE:    Zxx_RE  Zxx_IM  Zxy_RE  Zxy_IM
    - w1_crossHy_IM:    Zxx_RE  Zxx_IM  Zxy_RE  Zxy_IM

    Predictors Ey

    - w1_crossHx_RE:    Zyx_RE  Zyx_IM  Zyy_RE  Zyy_IM
    - w1_crossHx_IM:    Zyx_RE  Zyx_IM  Zyy_RE  Zyy_IM
    - w1_crossHy_RE:    Zyx_RE  Zyx_IM  Zyy_RE  Zyy_IM
    - w1_crossHy_IM:    Zyx_RE  Zyx_IM  Zyy_RE  Zyy_IM

    Note that the predictors are the same regardless of the output channel,
    only the observations change.
    """

    def __init__(
        self,
        metadata: RegressionInputMetadata,
        tf: TransferFunction,
        freqs: List[float],
        obs: List[Dict[str, np.ndarray]],
        preds: List[np.ndarray],
    ):
        """
        Initialisation of regression input data

        Parameters
        ----------
        metadata : RegressionInputMetadata
            The metadata, mainly to hold the various processing histories that
            have been combined to produce the data
        tf : TransferFunction
            The transfer function that is to be solver
        freqs : List[float]
            The evaluation frequencies
        obs : List[Dict[str, np.ndarray]]
            The observations (output channels). This is a list of dictionaries.
            The entries into the list are on for each evaluation evaluation
            frequency. They keys of the dictionary are the output channels.
        preds : List[np.ndarray]
            The predictions, an entry for each evaluation frequenvy
        """
        self.metadata = metadata
        self.tf = tf
        self.freqs = freqs
        self.obs = obs
        self.preds = preds

    @property
    def n_freqs(self) -> int:
        """Get the number of frequencies"""
        return len(self.freqs)

    def get_inputs(self, freq_idx: int, out_chan: str) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get observations and predictions

        Parameters
        ----------
        freq_idx : int
            The evaluation frequency index
        out_chan : str
            The output channel

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Observations and predictons
        """
        return self.obs[freq_idx][out_chan], self.preds[freq_idx]


class RegressionPreparerSpectra(ResisticsProcess):
    """
    Prepare regression data directly from spectra data

    This can be useful for running a single measurement
    """

    def run(self, tf: TransferFunction, spec_data: SpectraData) -> RegressionInputData:
        """Construct the linear equation for solving"""
        freqs = []
        obs = []
        preds = []
        for ilevel in range(spec_data.metadata.n_levels):
            level_metadata = spec_data.metadata.levels_metadata[ilevel]
            out_powers, in_powers = self._get_cross_powers(tf, spec_data, ilevel)
            for idx, freq in enumerate(level_metadata.freqs):
                logger.info(
                    f"Preparing regression data: level {ilevel}, freq. {idx} = {freq}"
                )
                freqs.append(freq)
                obs.append(self._get_obs(tf, out_powers[..., idx]))
                preds.append(self._get_preds(tf, in_powers[..., idx]))
        record = self._get_record("Produced regression input data for spectra data")
        metadata = RegressionInputMetadata(contributors={"data": spec_data.metadata})
        metadata.history.add_record(record)
        return RegressionInputData(metadata, tf, freqs, obs, preds)

    def _get_cross_powers(
        self, tf: TransferFunction, spec_data: SpectraData, level: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get cross powers

        Spectra data is:

        [n_wins, n_chans, n_freqs]

        To multiply each in/out channel with the cross channels, broadcasting is
        used. Using output channels as an example, this is what we have:

        out_data = [n_wins, n_out_chans, n_freqs]
        cross_data = [n_wins, n_cross_chans, n_freqs]

        The aim is to achieve an array that looks like this:

        cross_powers = [n_wins, n_out_chans, n_cross_chans, n_freqs]

        This can be achieved by numpy broadcasting the two arrays as follows

        out_data = [n_wins, n_out_chans, new_axis, n_freqs]
        cross_data = [n_wins, new_axis, n_cross_chans, n_freqs]

        Parameters
        ----------
        tf : TransferFunction
            Definition of transfer function
        spec_data : SpectraTimeData
            Spectra data for a decimation level
        level : int
            The decimation level

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Cross powers with output channels and cross powers with input
            channels
        """
        # prepare to calculate the crosspowers
        out_data = spec_data.get_chans(level, tf.out_chans)
        in_data = spec_data.get_chans(level, tf.in_chans)
        cross_data = spec_data.get_chans(level, tf.cross_chans)
        cross_data = np.conj(cross_data[:, np.newaxis, ...])

        # multiply using broadcasting
        out_powers = out_data[..., np.newaxis, :] * cross_data
        in_powers = in_data[..., np.newaxis, :] * cross_data
        return out_powers, in_powers

    def _get_obs(
        self, tf: TransferFunction, out_powers: np.ndarray
    ) -> Dict[str, np.ndarray]:
        """
        Get observations for an output channel

        Parameters
        ----------
        tf : TransferFunction
            Definition of transfer function
        out_powers : np.ndarray
            The cross powers for the output channels

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary with output channel as key and observations as value
        """
        obs = {}
        for idx, out_chan in enumerate(tf.out_chans):
            flattened = out_powers[:, idx, ...].flatten()
            obs_chan = np.empty((flattened.size * 2), dtype=float)
            obs_chan[0::2] = flattened.real
            obs_chan[1::2] = flattened.imag
            obs[out_chan] = obs_chan
        return obs

    def _get_obs_chan(self, tf: TransferFunction, out_powers: np.ndarray) -> np.ndarray:
        """
        Get observations for a single output channel

        The data comes in as:

        [n_wins, n_cross_chans]

        The aim is to turn this into an array of obserations for this output
        channel. The shape should be:

        [n_wins * n_cross_chans]

        Next, make everything floats to allow use of a wider range of open
        source solvers. This means interleaving the real and imaginary
        components into a longer single array, giving a final shape of:

        [n_wins * n_cross_chans * 2 ]

        Considering a concrete example for two windows:

        [[<Ex1, Hx1>, <Ex1,Hy1>], [[<Ex2, Hx2>, <Ex2,Hy2>]]

        Should become:

        [<Ex1, Hx1>.Real, <Ex1, Hx1>.Imag, <Ex1,Hy1>.Real, <Ex1,Hy1>.Imag,
        <Ex2, Hx2>.Real, <Ex2, Hx2>.Imag, <Ex2,Hy2>.Real, <Ex2,Hy2>.Imag]

        Parameters
        ----------
        tf : TransferFunction
            The transfer function definition
        out_powers : np.ndarray
            The cross powers for the a single output channel

        Returns
        -------
        np.ndarray
            The observations as a float
        """
        flattened = out_powers.flatten()
        obs = np.empty((flattened.size * 2), dtype=float)
        obs[0::2] = flattened.real
        obs[1::2] = flattened.imag
        return obs

    def _get_preds(self, tf: TransferFunction, in_powers: np.ndarray) -> np.ndarray:
        """
        Construct the predictors

        The in_powers is received with shape

        [n_wins, n_in_chans, n_cross_chans]

        The aim is to make this into

        [n_wins * n_cross_chans * 2, n_in_chans * 2]

        Parameters
        ----------
        tf : TransferFunction
            Transfer function definition
        in_powers : np.ndarray
            The cross powers for the input channels

        Returns
        -------
        np.ndarray
            The predictors
        """
        np.swapaxes(in_powers, 1, 2)
        n_wins = in_powers.shape[0]
        entries_per_win = tf.n_cross * 2
        preds = np.empty((n_wins * entries_per_win, tf.n_in * 2), dtype=float)
        for iwin in range(0, n_wins):
            idx_from = iwin * entries_per_win
            idx_to = idx_from + entries_per_win
            preds[idx_from:idx_to, :] = self._get_preds_win(tf, in_powers[iwin])
        return preds

    def _get_preds_win(self, tf: TransferFunction, in_powers: np.ndarray) -> np.ndarray:
        """Get predictors for a window"""
        preds_win = np.empty((tf.n_cross * 2, tf.n_in * 2), dtype=float)
        in_powers = np.swapaxes(in_powers, 0, 1)
        in_real = np.real(in_powers)
        in_imag = np.imag(in_powers)
        preds_win[0::2, 0::2] = in_real
        preds_win[0::2, 1::2] = -in_imag
        preds_win[1::2, 0::2] = in_imag
        preds_win[1::2, 1::2] = in_real
        return preds_win


class RegressionPreparerGathered(ResisticsProcess):
    """
    Regression preparer for gathered data

    In nearly all cases, this is the regresson preparer to use. As input, it
    requires GatheredData.
    """

    def run(
        self, tf: TransferFunction, gathered_data: GatheredData
    ) -> RegressionInputData:
        """
        Create the RegressionInputData

        Parameters
        ----------
        tf : TransferFunction
            The transfer function
        gathered_data : GatheredData
            The gathered data

        Returns
        -------
        RegressionInputData
            Data that can be used as input into a solver
        """
        logger.info("Preparing regression data")
        logger.info(f"Out chans site: {gathered_data.out_data.metadata.site_name}")
        logger.info(f"Out chans: {gathered_data.out_data.metadata.chans}")
        logger.info(f"In chans site: {gathered_data.in_data.metadata.site_name}")
        logger.info(f"In chans: {gathered_data.in_data.metadata.chans}")
        logger.info(f"Cross chans site: {gathered_data.cross_data.metadata.site_name}")
        logger.info(f"Cross chans: {gathered_data.cross_data.metadata.chans}")
        return self._get_regression_data(tf, gathered_data)

    def _get_regression_data(
        self, tf: TransferFunction, gathered_data: GatheredData
    ) -> RegressionInputData:
        """
        Get the regression input data

        Parameters
        ----------
        tf : TransferFunction
            The transfer function
        gathered_data : GatheredData
            The gathered data

        Returns
        -------
        RegressionInputData
            Data to be used as input to a solver
        """
        freqs = []
        obs = []
        preds = []
        metadata = gathered_data.out_data.metadata
        logger.info(f"Preparing regression data for {metadata.n_evals} frequencies")
        for idx, freq in enumerate(tqdm(metadata.eval_freqs)):
            out_powers, in_powers = self._get_cross_powers(tf, gathered_data, idx)
            freqs.append(freq)
            obs_freq = self._get_obs(tf, out_powers)
            preds_freq = self._get_preds(tf, in_powers)
            obs.append(obs_freq)
            preds.append(preds_freq)
        record = self._get_record(
            f"Produced regression input data for {metadata.n_evals} frequencies"
        )
        metadata = RegressionInputMetadata(
            contributors={
                "out_data": gathered_data.out_data.metadata,
                "in_data": gathered_data.in_data.metadata,
                "cross_data": gathered_data.cross_data.metadata,
            }
        )
        metadata.history.add_record(record)
        return RegressionInputData(metadata, tf, freqs, obs, preds)

    def _get_cross_powers(
        self, tf: TransferFunction, gathered_data: GatheredData, eval_idx: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get cross powers

        Gathered data for an evaluation frequency is:

        [n_wins, n_chans]

        To multiply each in/out channel with the cross channels, broadcasting is
        used. Using output channels as an example, this is what we have:

        out_data = [n_wins, n_out_chans]
        cross_data = [n_wins, n_cross_chans]

        The aim is to achieve an array that looks like this:

        cross_powers = [n_wins, n_out_chans, n_cross_chans]

        This can be achieved by numpy broadcasting the two arrays as follows

        out_data = [n_wins, n_out_chans, new_axis]
        cross_data = [n_wins, new_axis, n_cross_chans]

        Parameters
        ----------
        tf : TransferFunction
            Definition of transfer function
        gathered_data : GatheredData
            All the gathered data
        eval_idx : int
            The evaluation frequency index

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Cross powers with output channels and cross powers with input
            channels
        """
        # calculate the cross powers
        out_data = gathered_data.out_data.data[eval_idx]
        in_data = gathered_data.in_data.data[eval_idx]
        cross_data = gathered_data.cross_data.data[eval_idx]
        cross_data = np.conj(cross_data[:, np.newaxis, ...])

        # multiply using broadcasting
        out_powers = out_data[..., np.newaxis] * cross_data
        in_powers = in_data[..., np.newaxis] * cross_data
        return out_powers, in_powers

    def _get_obs(
        self, tf: TransferFunction, out_powers: np.ndarray
    ) -> Dict[str, np.ndarray]:
        """
        Get observations for an output channel

        Parameters
        ----------
        tf : TransferFunction
            Definition of transfer function
        out_powers : np.ndarray
            The cross powers for the output channels

        Returns
        -------
        Dict[str, np.ndarray]
            Dictionary with output channel as key and observations as value
        """
        obs = {}
        for idx, out_chan in enumerate(tf.out_chans):
            flattened = out_powers[:, idx, ...].flatten()
            obs_chan = np.empty((flattened.size * 2), dtype=float)
            obs_chan[0::2] = flattened.real
            obs_chan[1::2] = flattened.imag
            obs[out_chan] = obs_chan
        return obs

    def _get_obs_chan(self, tf: TransferFunction, out_powers: np.ndarray) -> np.ndarray:
        """
        Get observations for a single output channel

        The data comes in as:

        [n_wins, n_cross_chans]

        The aim is to turn this into an array of obserations for this output
        channel. The shape should be:

        [n_wins * n_cross_chans]

        Next, make everything floats to allow use of a wider range of open
        source solvers. This means interleaving the real and imaginary
        components into a longer single array, giving a final shape of:

        [n_wins * n_cross_chans * 2 ]

        Considering a concrete example for two windows:

        [[<Ex1, Hx1>, <Ex1,Hy1>], [[<Ex2, Hx2>, <Ex2,Hy2>]]

        Should become:

        [<Ex1, Hx1>.Real, <Ex1, Hx1>.Imag, <Ex1,Hy1>.Real, <Ex1,Hy1>.Imag,
        <Ex2, Hx2>.Real, <Ex2, Hx2>.Imag, <Ex2,Hy2>.Real, <Ex2,Hy2>.Imag]

        Parameters
        ----------
        tf : TransferFunction
            The transfer function definition
        out_powers : np.ndarray
            The cross powers for the a single output channel

        Returns
        -------
        np.ndarray
            The observations as a float
        """
        flattened = out_powers.flatten()
        obs = np.empty((flattened.size * 2), dtype=float)
        obs[0::2] = flattened.real
        obs[1::2] = flattened.imag
        return obs

    def _get_preds(self, tf: TransferFunction, in_powers: np.ndarray) -> np.ndarray:
        """
        Construct the predictors

        The in_powers is received with shape

        [n_wins, n_in_chans, n_cross_chans]

        The aim is to make this into

        [n_wins * n_cross_chans * 2, n_in_chans * 2]

        Parameters
        ----------
        tf : TransferFunction
            Transfer function definition
        in_powers : np.ndarray
            The cross powers for the input channels

        Returns
        -------
        np.ndarray
            The predictors
        """
        np.swapaxes(in_powers, 1, 2)
        n_wins = in_powers.shape[0]
        entries_per_win = tf.n_cross * 2
        preds = np.empty((n_wins * entries_per_win, tf.n_in * 2), dtype=float)
        for iwin in range(0, n_wins):
            idx_from = iwin * entries_per_win
            idx_to = idx_from + entries_per_win
            preds[idx_from:idx_to, :] = self._get_preds_win(tf, in_powers[iwin])
        return preds

    def _get_preds_win(self, tf: TransferFunction, in_powers: np.ndarray) -> np.ndarray:
        """Get predictors for a window"""
        preds_win = np.empty((tf.n_cross * 2, tf.n_in * 2), dtype=float)
        in_powers = np.swapaxes(in_powers, 0, 1)
        in_real = np.real(in_powers)
        in_imag = np.imag(in_powers)
        preds_win[0::2, 0::2] = in_real
        preds_win[0::2, 1::2] = -in_imag
        preds_win[1::2, 0::2] = in_imag
        preds_win[1::2, 1::2] = in_real
        return preds_win


class Solution(WriteableMetadata):
    """
    Class to hold a transfer function solution

    Examples
    --------
    >>> from resistics.testing import solution_mt
    >>> solution = solution_mt()
    >>> print(solution.tf.to_string())
    | Ex | = | Ex_Hx Ex_Hy | | Hx |
    | Ey |   | Ey_Hx Ey_Hy | | Hy |
    >>> solution.n_freqs
    5
    >>> solution.freqs
    [10.0, 20.0, 30.0, 40.0, 50.0]
    >>> solution.periods.tolist()
    [0.1, 0.05, 0.03333333333333333, 0.025, 0.02]
    >>> solution.components["ExHx"]
    Component(real=[1.0, 1.0, 2.0, 2.0, 3.0], imag=[5.0, 5.0, 4.0, 4.0, 3.0])
    >>> solution.components["ExHy"]
    Component(real=[1.0, 2.0, 3.0, 4.0, 5.0], imag=[-5.0, -4.0, -3.0, -2.0, -1.0])

    To get the components as an array, either get_component or subscripting
    be used

    >>> solution["ExHy"]
    array([1.-5.j, 2.-4.j, 3.-3.j, 4.-2.j, 5.-1.j])
    >>> solution["ab"]
    Traceback (most recent call last):
    ...
    ValueError: Component ab not found in ['ExHx', 'ExHy', 'EyHx', 'EyHy']

    It is also possible to get the tensor values at a particular evaluation
    frequency

    >>> solution.get_tensor(2)
    array([[ 2.+4.j,  3.-3.j],
           [-3.+3.j, -2.-4.j]])
    """

    tf: TransferFunction
    """The transfer function that was solved"""
    freqs: List[float]
    """The evaluation frequencies"""
    components: Dict[str, Component]
    """The solution"""
    history: History
    """The processing history"""
    contributors: Dict[str, Union[SiteCombinedMetadata, SpectraMetadata]]
    """The contributors to the solution with their respective details"""

    def __getitem__(self, key: str) -> np.ndarray:
        """
        Solution for a single component for all evaluation frequencies

        The arguments should be output channel followed by input channel

        Parameters
        ----------
        key : str
            The component key

        Returns
        -------
        np.ndarray
            The component values as an array

        Raises
        ------
        ValueError
            If incorrect number of arguments
        """
        if not isinstance(key, str):
            raise ValueError("Subscripting takes only 1 argument != {len(arg)}")
        return self.get_component(key)

    @property
    def n_freqs(self):
        """Get the number of evaluation frequencies"""
        return len(self.freqs)

    @property
    def periods(self) -> np.ndarray:
        """Get the periods"""
        return np.reciprocal(self.freqs)

    def get_component(self, key: str) -> np.ndarray:
        """
        Get the solution for a single component for all the evaluation
        frequencies

        Parameters
        ----------
        key : str
            The component key

        Returns
        -------
        np.ndarray
            The component data in an array

        Raises
        ------
        ValueError
            If the component does not exist in the solution
        """
        if key not in self.components:
            raise ValueError(
                f"Component {key} not found in {list(self.components.keys())}"
            )
        return self.components[key].to_numpy()

    def get_tensor(self, eval_idx: int) -> np.ndarray:
        """
        Get the tensor at a single evaluation frequency

        Parameters
        ----------
        eval_idx : int
            The index of the evaluation frequency

        Returns
        -------
        np.ndarray
            The tensor as a numpy array
        """
        tensor = np.zeros(shape=(self.tf.n_out, self.tf.n_in), dtype=np.complex128)
        for out_idx, out_chan in enumerate(self.tf.out_chans):
            for in_idx, in_chan in enumerate(self.tf.in_chans):
                key = get_component_key(out_chan, in_chan)
                tensor[out_idx, in_idx] = self.components[key].get_value(eval_idx)
        return tensor


class Solver(ResisticsProcess):
    """General resistics solver"""

    def run(self, regression_input: RegressionInputData) -> Solution:
        """Every solver should have a run method"""
        raise NotImplementedError("Run not implemented in parent Solver class")


class SolverScikit(Solver):
    """Base class for Scikit learn solvers"""

    fit_intercept: bool = False
    """Flag for adding an intercept term"""
    normalize: bool = False
    """Flag for normalizing, only used if fit_intercept is True"""

    def _solve(
        self, regression_input: RegressionInputData, model: BaseEstimator
    ) -> Solution:
        """
        Get the regression solution for all evaluation frequencies

        Parameters
        ----------
        regression_input : RegressionInputData
            The regression input data
        model : BaseEstimator
            The model to use to solve the linear regressions

        Returns
        -------
        Solution
            The solution for the transfer function
        """
        n_freqs = regression_input.n_freqs
        tf = regression_input.tf
        tensors = np.ndarray((n_freqs, tf.n_out, tf.n_in), dtype=np.complex128)
        logger.info(f"Solving for {n_freqs} evaluation frequencies")
        for eval_idx in tqdm(range(n_freqs)):
            for iout, out_chan in enumerate(tf.out_chans):
                obs, preds = regression_input.get_inputs(eval_idx, out_chan)
                coef = self._get_coef(model, obs, preds)
                tensors[eval_idx, iout] = self._get_tensor(tf, coef)
        return self._get_solution(tf, regression_input, tensors)

    def _get_coef(
        self, model: BaseEstimator, obs: np.ndarray, preds: np.ndarray
    ) -> np.ndarray:
        """
        Get coefficients for a single evaluation frequency and output channel

        Parameters
        ----------
        model : BaseEstimator
            sklearn base estimator
        obs : np.ndarray
            The observations
        preds : np.ndarray
            The predictors

        Returns
        -------
        np.ndarray
            The coefficients
        """
        model.fit(preds, obs)
        return model.coef_

    def _get_tensor(self, tf: TransferFunction, coef: np.ndarray) -> np.ndarray:
        """
        Rearrange the coefficients into a tensor

        Recall that the real and complex parts of the problem are separated out
        to allow use of a greater selection of solvers. Therefore, part of
        the job of this function is to reform the complex numbers.

        Parameters
        ----------
        tf : TransferFunction
            The transfer fuction
        coef : np.ndarray
            The coefficients

        Returns
        -------
        np.ndarray
            The coefficients in a tensor
        """
        values = np.empty((tf.n_in), dtype=np.complex128)
        for in_idx in range(tf.n_in):
            idx_coef = in_idx * 2
            values[in_idx] = coef[idx_coef] + 1j * coef[idx_coef + 1]
        return values

    def _get_solution(
        self,
        tf: TransferFunction,
        regression_input: RegressionInputData,
        tensors: np.ndarray,
    ) -> Solution:
        """
        Get the solution

        Parameters
        ----------
        tf : TransferFunction
            The transfer function
        regression_input : RegressionInputData
            The regression input data
        tensors : np.ndarray
            The coefficients

        Returns
        -------
        Solution
            The transfer function solution
        """
        components = {}
        for out_idx, out_chan in enumerate(tf.out_chans):
            for in_idx, in_chan in enumerate(tf.in_chans):
                key = get_component_key(out_chan, in_chan)
                values = tensors[:, out_idx, in_idx]
                components[key] = Component(
                    real=values.real.tolist(), imag=values.imag.tolist()
                )
        history = History(**regression_input.metadata.history.dict())
        message = f"Solved {len(regression_input.freqs)} evaluation frequencies"
        history.add_record(self._get_record(message))
        return Solution(
            tf=tf,
            freqs=regression_input.freqs,
            components=components,
            history=history,
            contributors=regression_input.metadata.contributors,
        )


class SolverScikitOLS(SolverScikit):
    """
    Ordinary least squares solver

    This is simply a wrapper around the scikit learn least squares regression
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
    """

    n_jobs: int = -2
    """Number of jobs to run"""

    def run(self, regression_input: RegressionInputData) -> Solution:
        """Run ordinary least squares regression on the RegressionInputData"""
        from sklearn.linear_model import LinearRegression

        model = LinearRegression(
            fit_intercept=self.fit_intercept,
            normalize=self.normalize,
            n_jobs=self.n_jobs,
        )
        return self._solve(regression_input, model)


class SolverScikitHuber(SolverScikit):
    """
    Scikit Huber solver

    This is simply a wrapper around the scikit learn Huber Regressor. For
    more information, please see
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.HuberRegressor.html
    """

    epsilon: float = 1
    """The smaller the epsilon, the more robust it is to outliers."""

    def run(self, regression_input: RegressionInputData) -> Solution:
        """Run Huber Regressor regression on the RegressionInputData"""
        from sklearn.linear_model import HuberRegressor

        model = HuberRegressor(epsilon=self.epsilon)
        return self._solve(regression_input, model)


class SolverScikitTheilSen(SolverScikit):
    """
    Scikit Theil Sen solver

    This is simply a wrapper around the scikit learn Theil Sen Regressor. For
    more information, please see
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TheilSenRegressor.html
    """

    n_jobs: int = -2
    """Number of jobs to run"""
    max_subpopulation: int = 2_000
    """Maximum population. Reduce this if the process is taking a long time"""
    n_subsamples: Optional[int] = None
    """Number of rows to use for each solution"""

    def run(self, regression_input: RegressionInputData) -> Solution:
        """Run TheilSen regression on the RegressionInputData"""
        from sklearn.linear_model import TheilSenRegressor

        model = TheilSenRegressor(
            fit_intercept=self.fit_intercept,
            max_subpopulation=self.max_subpopulation,
            n_subsamples=self.n_subsamples,
            n_jobs=self.n_jobs,
        )
        return self._solve(regression_input, model)


class SolverScikitRANSAC(SolverScikit):
    """
    Run a RANSAC solver with LinearRegression as Base Estimator

    This is a wrapper around the scikit learn RANSAC regressor. More information
    can be found here
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RANSACRegressor.html
    """

    min_samples: float = 0.1
    """Minimum number of samples in each solution as a proportion of total"""
    max_trials: int = 1000
    """The maximum number of trials to run"""

    def run(self, regression_input: RegressionInputData) -> Solution:
        """Run RANSAC regression on the RegressionInputData"""
        from sklearn.linear_model import LinearRegression, RANSACRegressor

        model = RANSACRegressor(
            LinearRegression(
                fit_intercept=self.fit_intercept, normalize=self.normalize
            ),
            min_samples=self.min_samples,
            max_trials=self.max_trials,
        )
        return self._solve(regression_input, model)

    def _get_coef(
        self, model: BaseEstimator, obs: np.ndarray, preds: np.ndarray
    ) -> np.ndarray:
        """Get coefficients for single evaluation frequeny and output channel"""

        model.fit(preds, obs)
        return model.estimator_.coef_


class SolverScikitWLS(SolverScikitOLS):
    """
    Weighted least squares solver

    .. warning::

        This is homespun and is currently only experimental

    This is simply a wrapper around the scikit learn least squares regression
    using the sample_weight option
    https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html
    """

    n_jobs: int = -2
    """Number of jobs to run"""
    n_iter: int = 50

    def _get_coef(
        self, model: BaseEstimator, obs: np.ndarray, preds: np.ndarray
    ) -> np.ndarray:
        """
        Get coefficients for a single evaluation frequency and output channel

        Parameters
        ----------
        model : BaseEstimator
            sklearn base estimator
        obs : np.ndarray
            The observations
        preds : np.ndarray
            The predictors

        Returns
        -------
        np.ndarray
            The coefficients
        """
        from sklearn.preprocessing import RobustScaler

        weights = np.ones(shape=(obs.size))
        scalar = RobustScaler()
        iteration = 0
        while iteration < self.n_iter:
            model.fit(preds, obs, sample_weight=weights)
            obs_pred = model.predict(preds)
            resids = np.absolute(obs - obs_pred)
            resids_scaled = scalar.fit_transform(resids.reshape(-1, 1))
            weights = self.bisquare(resids_scaled)
            iteration += 1
        return model.coef_

    def bisquare(self, r: np.ndarray, k: float = 4.685) -> np.ndarray:
        """
        Bisquare location weights

        Parameters
        ----------
        r : np.ndarray
            Residuals
        k : float, None
            Tuning parameter. If None, a standard value will be used.

        Returns
        -------
        weights : np.ndarray
            The robust weights
        """
        r = r.reshape((r.shape[0]))
        ones = np.ones(shape=r.shape)
        thresh = np.minimum(ones, r / k)
        return np.power((1 - np.power(thresh, 2)), 2)

    def huber(self, r: np.ndarray, k: float = 1.345) -> np.ndarray:
        """Huber location weights

        Parameters
        ----------
        r : np.ndarray
            Residuals
        k : float
            Tuning parameter. If None, a standard value will be used.

        Returns
        -------
        weights : np.ndarray
            The robust weights
        """
        r = r.reshape((r.shape[0]))
        indices = np.where(r > k)
        weights = np.ones(shape=r.shape)
        weights[indices] = k / r[indices]
        return weights

    def trimmed_mean(self, r: np.ndarray, k: float = 2) -> np.ndarray:
        """
        Trimmed mean location weights

        Parameters
        ----------
        r : np.ndarray
            Residuals
        k : float
            Tuning parameter. If None, a standard value will be used.

        Returns
        -------
        weights : np.ndarray
            The robust weights
        """
        r = r.reshape((r.shape[0]))
        indices = np.where(r <= k)
        weights = np.zeros(shape=r.shape)
        weights[indices] = 1
        return weights.real



"""
Module containing functions and classes related to Spectra calculation and
manipulation

Spectra are calculated from the windowed, decimated time data. The inbuilt
Fourier transform implementation is inspired by the implementation of the
scipy stft function.
"""
from loguru import logger
from pathlib import Path
from typing import Union, Tuple, Dict, List, Any, Optional
from pydantic import PositiveInt
import numpy as np
import pandas as pd
import plotly.graph_objects as go

from resistics.common import ResisticsData, ResisticsProcess, History
from resistics.common import ResisticsWriter, Metadata, WriteableMetadata
from resistics.sampling import HighResDateTime
from resistics.time import ChanMetadata
from resistics.decimate import DecimationParameters
from resistics.window import WindowedData, WindowedLevelMetadata


class SpectraLevelMetadata(Metadata):
    """Metadata for spectra of a windowed decimation level"""

    fs: float
    """The sampling frequency of the decimation level"""
    n_wins: int
    """The number of windows"""
    win_size: PositiveInt
    """The window size in samples"""
    olap_size: PositiveInt
    """The overlap size in samples"""
    index_offset: int
    """The global window offset for local window 0"""
    n_freqs: int
    """The number of frequencies in the frequency data"""
    freqs: List[float]
    """List of frequencies"""

    @property
    def nyquist(self) -> float:
        """Get the nyquist frequency"""
        return self.fs / 2


class SpectraMetadata(WriteableMetadata):
    """Metadata for spectra data"""

    fs: List[float]
    chans: List[str]
    n_chans: Optional[int] = None
    n_levels: int
    first_time: HighResDateTime
    last_time: HighResDateTime
    system: str = ""
    serial: str = ""
    wgs84_latitude: float = -999.0
    wgs84_longitude: float = -999.0
    easting: float = -999.0
    northing: float = -999.0
    elevation: float = -999.0
    chans_metadata: Dict[str, ChanMetadata]
    levels_metadata: List[SpectraLevelMetadata]
    ref_time: HighResDateTime
    history: History = History()

    class Config:

        extra = "ignore"


class SpectraData(ResisticsData):
    """
    Class for holding spectra data

    The spectra data is stored in the class as a dictionary mapping decimation
    level to numpy array. The shape of the array for each decimation level is:

    n_wins x n_chans x n_freqs
    """

    def __init__(self, metadata: SpectraMetadata, data: Dict[int, np.ndarray]):
        """
        Initialise spectra data

        Parameters
        ----------
        metadata : SpectraMetadata
            Metadata for the spectra data
        data : Dict[int, np.ndarray]
            Dictionary of data, one entry for each evaluation level
        """
        logger.debug(f"Creating SpectraData with data type {data[0].dtype}")
        self.metadata = metadata
        self.data = data

    def get_level(self, level: int) -> np.ndarray:
        """Get the spectra data for a decimation level"""
        if level >= self.metadata.n_levels:
            raise ValueError(f"Level {level} not <= max {self.metadata.n_levels - 1}")
        return self.data[level]

    def get_chan(self, level: int, chan: str) -> np.ndarray:
        """Get the channel spectra data for a decimation level"""
        from resistics.errors import ChannelNotFoundError

        if chan not in self.metadata.chans:
            raise ChannelNotFoundError(chan, self.metadata.chans)
        idx = self.metadata.chans.index(chan)
        return self.data[level][..., idx, :]

    def get_chans(self, level: int, chans: List[str]) -> np.ndarray:
        """Get the channels spectra data for a decimation level"""
        from resistics.errors import ChannelNotFoundError

        for chan in chans:
            if chan not in self.metadata.chans:
                raise ChannelNotFoundError(chan, self.metadata.chans)
        indices = [self.metadata.chans.index(chan) for chan in chans]
        return self.data[level][..., indices, :]

    def get_freq(self, level: int, idx: int) -> np.ndarray:
        """Get the spectra data at a frequency index for a decimation level"""
        n_freqs = self.metadata.levels_metadata[level].n_freqs
        if idx < 0 or idx >= n_freqs:
            raise ValueError(f"Freq. index {idx} not 0 <= idx < {n_freqs}")
        return np.squeeze(self.data[level][..., idx])

    def get_mag_phs(
        self, level: int, unwrap: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Get magnitude and phase for a decimation level"""
        spec = self.data[level]
        if unwrap:
            return np.absolute(spec), np.unwrap(np.angle(spec))
        return np.absolute(spec), np.angle(spec)

    def get_timestamps(self, level: int) -> pd.DatetimeIndex:
        """
        Get the start time of each window

        Note that this does not use high resolution timestamps

        Parameters
        ----------
        level : int
            The decimation level

        Returns
        -------
        pd.DatetimeIndex
            The starts of each window

        Raises
        ------
        ValueError
            If the level is out of range
        """
        from resistics.window import get_win_starts

        if level >= self.metadata.n_levels:
            raise ValueError(f"Level {level} not <= max {self.metadata.n_levels - 1}")
        level_metadata = self.metadata.levels_metadata[level]
        return get_win_starts(
            self.metadata.ref_time,
            level_metadata.win_size,
            level_metadata.olap_size,
            level_metadata.fs,
            level_metadata.n_wins,
            level_metadata.index_offset,
        )

    def plot_stack(
        self,
        level: int,
        max_pts: int = 10_000,
        grouping: str = "12h",
        offset: str = "0h",
    ) -> go.Figure:
        """
        Stack the spectra with optional time grouping

        Parameters
        ----------
        level : int
            The decimation level
        max_pts : int, optional
            The maximum number of points in any individual plot before applying
            lttbc downsampling, by default 10_000
        grouping : str, optional
            A grouping interval as a pandas freq string, by default "6h"
        offset : str, optional
            A time offset to add to the grouping, by default "0h". For instance,
            to plot night time and day time spectra, set grouping to "12h" and
            offset to "6h"

        Returns
        -------
        go.Figure
            The plotly figure
        """
        from resistics.plot import PlotData1D, figure_columns_as_lines, plot_columns_1d

        if grouping is None:
            first_date = pd.Timestamp(self.metadata.first_time.isoformat()).floor("D")
            last_date = pd.Timestamp(self.metadata.last_time.isoformat()).ceil("D")
            grouping = last_date - first_date
        level_metadata = self.metadata.levels_metadata[level]
        df = pd.DataFrame(
            data=np.arange(level_metadata.n_wins),
            index=self.get_timestamps(level),
            columns=["local"],
        )
        # prepare the plot
        subplots = self.metadata.chans
        subplot_columns = {x: [x] for x in subplots}
        y_labels = {x: "Magnitude" for x in subplots}
        fig = figure_columns_as_lines(
            subplots=subplots, y_labels=y_labels, x_label="Frequency"
        )
        fig.update_yaxes(type="log")
        # group by the grouping frequency, iterate over the groups and plot
        for idx, group in df.groupby(pd.Grouper(freq=grouping, offset=offset)):
            stack = np.mean(np.absolute(self.data[level][group["local"]]), axis=0)
            plot_data = PlotData1D(
                x=np.array(level_metadata.freqs), data=stack, rows=self.metadata.chans
            )
            plot_columns_1d(
                fig,
                plot_data,
                subplots,
                subplot_columns,
                max_pts=max_pts,
                label_prefix=str(idx),
            )
        return fig

    def plot_section(self, level: int, grouping="30T") -> go.Figure:
        """
        Plot a spectra section

        Parameters
        ----------
        level : int
            The decimation level to plot
        grouping : str, optional
            The time domain resolution, by default "30T"

        Returns
        -------
        go.Figure
            A plotly figure
        """
        from resistics.plot import figure_columns_as_lines

        level_metadata = self.metadata.levels_metadata[level]
        df = pd.DataFrame(
            data=np.arange(level_metadata.n_wins),
            index=self.get_timestamps(level),
            columns=["local"],
        )
        # create the figure
        subplots = self.metadata.chans
        y_labels = {x: "Frequency Hz" for x in subplots}
        fig = figure_columns_as_lines(
            subplots=subplots, y_labels=y_labels, x_label="Date"
        )
        colorbar_len = 0.90 / self.metadata.n_chans
        colorbar_inc = 0.83 / (self.metadata.n_chans - 1)
        # group by the grouping frequency, iterate over the groups and plot
        data = {}
        for idx, group in df.groupby(pd.Grouper(freq=grouping)):
            data[idx] = np.mean(np.absolute(self.data[level][group["local"]]), axis=0)
        for idx, chan in enumerate(self.metadata.chans):
            df_data = pd.DataFrame(
                data={k: v[idx] for k, v in data.items()}, index=level_metadata.freqs
            )
            z = np.log10(df_data.values)
            z_min = np.ceil(z.min())
            z_max = np.floor(z.max())
            z_range = np.arange(z_min, z_max + 1)
            colorbar = dict(
                title=f"Amplitude {chan}",
                tickvals=z_range,
                ticktext=[f"10^{int(x)}" for x in z_range],
                y=0.92 - idx * colorbar_inc,
                len=colorbar_len,
            )
            heatmap = go.Heatmap(
                z=z,
                x=pd.to_datetime(df_data.columns) + pd.Timedelta(grouping) / 2,
                y=df_data.index,
                zmin=z_min,
                zmax=z_max,
                colorscale="viridis",
                colorbar=colorbar,
            )
            fig.append_trace(heatmap, row=idx + 1, col=1)
        return fig


class FourierTransform(ResisticsProcess):
    """
    Perform a Fourier transform of the windowed data

    The processor is inspired by the scipy.signal.stft function which performs
    a similar process and involves a Fourier transform along the last axis of
    the windowed data.

    Parameters
    ----------
    win_fnc : Union[str, Tuple[str, float]]
        The window to use before performing the FFT, by default ("kaiser", 14)
    detrend : Union[str, None]
        Type of detrending to apply before performing FFT, by default linear
        detrend. Setting to None will not apply any detrending to the data prior
        to the FFT
    workers : int
        The number of CPUs to use, by default max - 2

    Examples
    --------
    This example will get periodic decimated data, perfrom windowing and run the
    Fourier transform on the windowed data.

    .. plot::
        :width: 90%

        >>> import matplotlib.pyplot as plt
        >>> import numpy as np
        >>> from resistics.testing import decimated_data_periodic
        >>> from resistics.window import WindowSetup, Windower
        >>> from resistics.spectra import FourierTransform
        >>> frequencies = {"chan1": [870, 590, 110, 32, 12], "chan2": [480, 375, 210, 60, 45]}
        >>> dec_data = decimated_data_periodic(frequencies, fs=128)
        >>> dec_data.metadata.chans
        ['chan1', 'chan2']
        >>> print(dec_data.to_string())
        <class 'resistics.decimate.DecimatedData'>
                   fs        dt  n_samples           first_time                        last_time
        level
        0      2048.0  0.000488      16384  2021-01-01 00:00:00  2021-01-01 00:00:07.99951171875
        1       512.0  0.001953       4096  2021-01-01 00:00:00    2021-01-01 00:00:07.998046875
        2       128.0  0.007812       1024  2021-01-01 00:00:00      2021-01-01 00:00:07.9921875

        Perform the windowing

        >>> win_params = WindowSetup().run(dec_data.metadata.n_levels, dec_data.metadata.fs)
        >>> win_data = Windower().run(dec_data.metadata.first_time, win_params, dec_data)

        And then the Fourier transform. By default, the data will be (linearly)
        detrended and mutliplied by a Kaiser window prior to the Fourier
        transform

        >>> spec_data = FourierTransform().run(win_data)

        For plotting of magnitude, let's stack the spectra

        >>> freqs_0 = spec_data.metadata.levels_metadata[0].freqs
        >>> data_0 = np.absolute(spec_data.data[0]).mean(axis=0)
        >>> freqs_1 = spec_data.metadata.levels_metadata[1].freqs
        >>> data_1 = np.absolute(spec_data.data[1]).mean(axis=0)
        >>> freqs_2 = spec_data.metadata.levels_metadata[2].freqs
        >>> data_2 = np.absolute(spec_data.data[2]).mean(axis=0)

        Now plot

        >>> plt.subplot(3,1,1) # doctest: +SKIP
        >>> plt.plot(freqs_0, data_0[0], label="chan1") # doctest: +SKIP
        >>> plt.plot(freqs_0, data_0[1], label="chan2") # doctest: +SKIP
        >>> plt.grid()
        >>> plt.title("Decimation level 0") # doctest: +SKIP
        >>> plt.legend() # doctest: +SKIP
        >>> plt.subplot(3,1,2) # doctest: +SKIP
        >>> plt.plot(freqs_1, data_1[0], label="chan1") # doctest: +SKIP
        >>> plt.plot(freqs_1, data_1[1], label="chan2") # doctest: +SKIP
        >>> plt.grid()
        >>> plt.title("Decimation level 1") # doctest: +SKIP
        >>> plt.legend() # doctest: +SKIP
        >>> plt.subplot(3,1,3) # doctest: +SKIP
        >>> plt.plot(freqs_2, data_2[0], label="chan1") # doctest: +SKIP
        >>> plt.plot(freqs_2, data_2[1], label="chan2") # doctest: +SKIP
        >>> plt.grid()
        >>> plt.title("Decimation level 2") # doctest: +SKIP
        >>> plt.legend() # doctest: +SKIP
        >>> plt.xlabel("Frequency") # doctest: +SKIP
        >>> plt.tight_layout() # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
    """

    win_fnc: Union[str, Tuple[str, float]] = ("kaiser", 14)
    detrend: Union[str, None] = "linear"
    workers: int = -2

    def run(self, win_data: WindowedData) -> SpectraData:
        """
        Perform the FFT

        Data is padded to the next fast length before performing the FFT to
        speed up processing. Therefore, the output length may not be as
        expected.

        Parameters
        ----------
        win_data : WindowedData
            The input windowed data

        Returns
        -------
        SpectraData
            The Fourier transformed output
        """
        from scipy.fft import next_fast_len, rfftfreq

        metadata_dict = win_data.metadata.dict()
        data = {}
        spectra_levels_metadata = []
        messages = []
        logger.info("Performing fourier transforms of windowed decimated data")
        for ilevel in range(win_data.metadata.n_levels):
            logger.info(f"Transforming level {ilevel}")
            level_metadata = win_data.metadata.levels_metadata[ilevel]
            win_size = level_metadata.win_size
            n_transform = next_fast_len(win_size, real=True)
            logger.debug(f"Padding size {win_size} to next fast len {n_transform}")
            freqs = rfftfreq(n=n_transform, d=1.0 / level_metadata.fs).tolist()
            data[ilevel] = self._get_level_data(
                level_metadata, win_data.get_level(ilevel), n_transform
            )
            spectra_levels_metadata.append(
                self._get_level_metadata(level_metadata, freqs)
            )
            messages.append(f"Calculated spectra for level {ilevel}")
        metadata = self._get_metadata(metadata_dict, spectra_levels_metadata)
        metadata.history.add_record(self._get_record(messages))
        logger.info("Fourier transforms completed")
        return SpectraData(metadata, data)

    def _get_level_data(
        self, metadata: WindowedLevelMetadata, data: np.ndarray, n_transform: int
    ) -> np.ndarray:
        """
        Run the spectra calculation for a single decimation level

        The input is an array with shape:

        n_wins x n_chans x win_size

        And the output has shape

        n_wins x n_chans x n_transform

        Parameters
        ----------
        metadata : WindowedLevelMetadata
            Level metadata
        data : np.ndarray
            Data to transform
        n_transform : int
            Size of the transform

        Returns
        -------
        np.ndarray
            Transformed data for all windows
        """
        from scipy import signal
        from scipy.fft import rfft

        # detrend and apply window
        if self.detrend is not None:
            data = signal.detrend(data, axis=-1, type=self.detrend)
        win_coeffs = self._get_window(metadata.win_size).astype(data.dtype)
        data = data * win_coeffs
        # perform the fft on the last axis
        return rfft(data, n=n_transform, axis=-1, norm="ortho", workers=self.workers)

    def _get_window(self, win_size: int):
        """Get the window to apply to the data"""
        from scipy.signal import get_window
        from scipy.signal.windows import dpss

        if self.win_fnc == "dpss":
            return dpss(win_size, 5)
        return get_window(self.win_fnc, win_size)

    def _get_level_metadata(
        self, level_metadata: WindowedLevelMetadata, freqs: List[float]
    ) -> SpectraLevelMetadata:
        """Get the spectra metadata for a decimation level"""
        metadata_dict = level_metadata.dict()
        metadata_dict["n_freqs"] = len(freqs)
        metadata_dict["freqs"] = freqs
        return SpectraLevelMetadata(**metadata_dict)

    def _get_metadata(
        self,
        metadata_dict: Dict[str, Any],
        levels_metadata: List[SpectraLevelMetadata],
    ) -> SpectraMetadata:
        """Get the metadata for the windowed data"""
        metadata_dict.pop("file_info")
        metadata_dict["levels_metadata"] = levels_metadata
        return SpectraMetadata(**metadata_dict)


class EvaluationFreqs(ResisticsProcess):
    """
    Calculate the spectra values at the evaluation frequencies

    This is done using linear interpolation in the complex domain

    Example
    -------
    The example will show interpolation to evaluation frequencies on a very
    simple example. Begin by generating some example spectra data.

    >>> from resistics.decimate import DecimationSetup
    >>> from resistics.spectra import EvaluationFreqs
    >>> from resistics.testing import spectra_data_basic
    >>> spec_data = spectra_data_basic()
    >>> spec_data.metadata.n_levels
    1
    >>> spec_data.metadata.chans
    ['chan1']
    >>> spec_data.metadata.levels_metadata[0].summary()
    {
        'fs': 180.0,
        'n_wins': 2,
        'win_size': 20,
        'olap_size': 5,
        'index_offset': 0,
        'n_freqs': 10,
        'freqs': [0.0, 10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0]
    }

    The spectra data has only a single channel and a single level which has 2
    windows. Now define our evaluation frequencies.

    >>> eval_freqs = [1, 12, 23, 34, 45, 56, 67, 78, 89]
    >>> dec_setup = DecimationSetup(n_levels=1, per_level=9, eval_freqs=eval_freqs)
    >>> dec_params = dec_setup.run(spec_data.metadata.fs[0])
    >>> dec_params.summary()
    {
        'fs': 180.0,
        'n_levels': 1,
        'per_level': 9,
        'min_samples': 256,
        'eval_freqs': [1.0, 12.0, 23.0, 34.0, 45.0, 56.0, 67.0, 78.0, 89.0],
        'dec_factors': [1],
        'dec_increments': [1],
        'dec_fs': [180.0]
    }

    Now calculate the spectra at the evaluation frequencies

    >>> eval_data = EvaluationFreqs().run(dec_params, spec_data)
    >>> eval_data.metadata.levels_metadata[0].summary()
    {
        'fs': 180.0,
        'n_wins': 2,
        'win_size': 20,
        'olap_size': 5,
        'index_offset': 0,
        'n_freqs': 9,
        'freqs': [1.0, 12.0, 23.0, 34.0, 45.0, 56.0, 67.0, 78.0, 89.0]
    }

    To double check everything is as expected, let's compare the data. Comparing
    window 1 gives

    >>> print(spec_data.data[0][0, 0])
    [0.+0.j 1.+1.j 2.+2.j 3.+3.j 4.+4.j 5.+5.j 6.+6.j 7.+7.j 8.+8.j 9.+9.j]
    >>> print(eval_data.data[0][0, 0])
    [0.1+0.1j 1.2+1.2j 2.3+2.3j 3.4+3.4j 4.5+4.5j 5.6+5.6j 6.7+6.7j 7.8+7.8j
     8.9+8.9j]

    And window 2

    >>> print(spec_data.data[0][1, 0])
    [-1. +1.j  0. +2.j  1. +3.j  2. +4.j  3. +5.j  4. +6.j  5. +7.j  6. +8.j
      7. +9.j  8.+10.j]
    >>> print(eval_data.data[0][1, 0])
    [-0.9+1.1j  0.2+2.2j  1.3+3.3j  2.4+4.4j  3.5+5.5j  4.6+6.6j  5.7+7.7j
      6.8+8.8j  7.9+9.9j]
    """

    def run(
        self, dec_params: DecimationParameters, spec_data: SpectraData
    ) -> SpectraData:
        """
        Interpolate spectra data to the evaluation frequencies

        This is a simple linear interpolation.

        Parameters
        ----------
        dec_params : DecimationParameters
            The decimation parameters which have the evaluation frequencies for
            each decimation level
        spec_data : SpectraData
            The spectra data

        Returns
        -------
        SpectraData
            The spectra data at the evaluation frequencies
        """
        metadata_dict = spec_data.metadata.dict()
        data = {}
        spectra_levels_metadata = []
        messages = []
        for ilevel in range(spec_data.metadata.n_levels):
            logger.info(f"Reducing freqs to evaluation freqs for level {ilevel}")
            level_metadata = spec_data.metadata.levels_metadata[ilevel]
            freqs = np.array(level_metadata.freqs)
            eval_freqs = np.array(dec_params.get_eval_freqs(ilevel))
            data[ilevel] = self._get_level_data(
                freqs, spec_data.get_level(ilevel), eval_freqs
            )
            spectra_levels_metadata.append(
                self._get_level_metadata(level_metadata, eval_freqs)
            )
        messages.append("Spectra reduced to evaluation frequencies")
        metadata = self._get_metadata(metadata_dict, spectra_levels_metadata)
        metadata.history.add_record(self._get_record(messages))
        logger.info("Fourier coefficients calculated at evaluation frequencies")
        return SpectraData(metadata, data)

    def _get_level_data(
        self, freqs: np.ndarray, data: np.ndarray, eval_freqs: np.ndarray
    ) -> np.ndarray:
        """
        Interpolate the spectra data to the evaluation frequencies

        The input data for a level has shape:

        n_wins x n_chans x n_freqs

        The new output data will have size:

        n_wins x n_chans x n_eval_freqs

        This process is doing a linear interpolation. As this is complex data
        and numpy does not have an interpolation along axis option,
        interpolation is done manually.

        First the evaluation frequencies are interpolated to their indices given
        the current frequencies and indices.

        Then these float indices are used to do the interpolation.

        Parameters
        ----------
        freqs : np.ndarray
            The input data frequencies
        data : np.ndarray
            The input spectra data
        eval_freqs : List[float]
            The evaluation frequencies

        Returns
        -------
        np.ndarray
            Output level data
        """
        index = np.arange(len(freqs))
        eval_indices = np.interp(eval_freqs, freqs, index)
        floors = np.floor(eval_indices).astype(int)
        ceils = np.ceil(eval_indices).astype(int)
        # cast portions to preserve original data type
        # otherwise, can expand complex64 to complex128
        portions = (eval_indices - floors).astype(data.dtype)
        diffs = data[..., ceils] - data[..., floors]
        add = np.squeeze(diffs[..., np.newaxis, :] * portions, axis=-2)
        return data[..., floors] + add

    def _get_level_metadata(
        self, level_metadata: SpectraLevelMetadata, eval_freqs: np.ndarray
    ) -> SpectraLevelMetadata:
        """Get the metadata for the decimation level"""
        metadata_dict = level_metadata.dict()
        metadata_dict["n_freqs"] = len(eval_freqs)
        metadata_dict["freqs"] = eval_freqs.tolist()
        return SpectraLevelMetadata(**metadata_dict)

    def _get_metadata(
        self, metadata_dict: Dict[str, Any], levels_metadata: List[SpectraLevelMetadata]
    ) -> SpectraMetadata:
        """Get metadata for the dataset"""
        metadata_dict.pop("file_info")
        metadata_dict["levels_metadata"] = levels_metadata
        return SpectraMetadata(**metadata_dict)


class SpectraDataWriter(ResisticsWriter):
    """Writer of resistics spectra data"""

    def run(self, dir_path: Path, spec_data: SpectraData) -> None:
        """
        Write out SpectraData

        Parameters
        ----------
        dir_path : Path
            The directory path to write to
        spec_data : SpectraData
            Spectra data to write out

        Raises
        ------
        WriteError
            If unable to write to the directory
        """
        from resistics.errors import WriteError

        if not self._check_dir(dir_path):
            raise WriteError(dir_path, "Unable to write to directory, check logs")
        logger.info(f"Writing spectra data to {dir_path}")
        metadata_path = dir_path / "metadata.json"
        data_path = dir_path / "data"
        np.savez_compressed(data_path, **{str(x): y for x, y in spec_data.data.items()})
        metadata = spec_data.metadata.copy()
        metadata.history.add_record(self._get_record(dir_path, type(spec_data)))
        metadata.write(metadata_path)


class SpectraDataReader(ResisticsProcess):
    """Reader of resistics spectra data"""

    def run(
        self, dir_path: Path, metadata_only: bool = False
    ) -> Union[SpectraMetadata, SpectraData]:
        """
        Read SpectraData

        Parameters
        ----------
        dir_path : Path
            The directory path to read from
        metadata_only : bool, optional
            Flag for getting metadata only, by default False

        Returns
        -------
        Union[SpectraMetadata, SpectraData]
            The SpectraData or SpectraMetadata if metadata_only is True

        Raises
        ------
        ReadError
            If the directory does not exist
        """
        from resistics.errors import ReadError

        if not dir_path.exists():
            raise ReadError(dir_path, "Directory does not exist")
        logger.info(f"Reading spectra data from {dir_path}")
        metadata_path = dir_path / "metadata.json"
        metadata = SpectraMetadata.parse_file(metadata_path)
        if metadata_only:
            return metadata
        data_path = dir_path / "data.npz"
        npz_file = np.load(data_path)
        data = {int(level): npz_file[level] for level in npz_file.files}
        messages = [f"Spectra data read from {dir_path}"]
        metadata.history.add_record(self._get_record(messages))
        return SpectraData(metadata, data)


class SpectraProcess(ResisticsProcess):
    """Parent class for spectra processes"""

    def run(self, spec_data: SpectraData) -> SpectraData:
        """Run a spectra processor"""
        raise NotImplementedError("Run is not implemented in the parent SpectraProcess")


class SpectraSmootherUniform(SpectraProcess):
    """
    Smooth a spectra with a uniform filter

    For more information, please refer to:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.uniform_filter1d.html

    Examples
    --------
    Smooth a simple spectra data instance

    >>> from resistics.spectra import SpectraSmootherUniform
    >>> from resistics.testing import spectra_data_basic
    >>> spec_data = spectra_data_basic()
    >>> smooth_data = SpectraSmootherUniform(length_proportion=0.5).run(spec_data)

    Look at the results for the two windows

    >>> spec_data.data[0][0,0]
    array([0.+0.j, 1.+1.j, 2.+2.j, 3.+3.j, 4.+4.j, 5.+5.j, 6.+6.j, 7.+7.j,
           8.+8.j, 9.+9.j])
    >>> smooth_data.data[0][0,0]
    array([0.8+0.8j, 1.2+1.2j, 2. +2.j , 3. +3.j , 4. +4.j , 5. +5.j ,
           6. +6.j , 7. +7.j , 7.8+7.8j, 8.2+8.2j])
    """

    length_proportion: float = 0.1

    def run(self, spec_data: SpectraData) -> SpectraData:
        """
        Smooth spectra data with a uniform smoother

        Parameters
        ----------
        spec_data : SpectraData
            The input spectra data

        Returns
        -------
        SpectraData
            The output spectra data
        """
        import scipy.ndimage as ndimage

        data = {}
        logger.info("Smoothing frequencies with uniform filter")
        messages = ["Smoothing frequencies with uniform filter"]
        for ilevel in range(spec_data.metadata.n_levels):
            n_freqs = spec_data.metadata.levels_metadata[ilevel].n_freqs
            smooth_length = self._get_smooth_length(n_freqs)
            logger.debug(f"Smoothing level {ilevel} with num points {smooth_length}")
            data[ilevel] = ndimage.uniform_filter1d(
                spec_data.get_level(ilevel), smooth_length, axis=-1
            )
            messages.append(f"Smoothed level {ilevel} with num points {smooth_length}")
        metadata = SpectraMetadata(**spec_data.metadata.dict())
        metadata.history.add_record(self._get_record(messages))
        logger.info("Fourier coefficients calculated at evaluation frequencies")
        return SpectraData(metadata, data)

    def _get_smooth_length(self, data_size: int) -> int:
        """Get the smoothing length given the size of the data"""
        length = int(self.length_proportion * data_size)
        if length % 2 == 0:
            length += 1
        if length < 1:
            return 1
        return length


class SpectraSmootherGaussian(SpectraProcess):
    """
    Smooth a spectra with a gaussian filter

    For more information, please refer to:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.gaussian_filter1d.html

    Examples
    --------
    Smooth a simple spectra data instance

    >>> from resistics.spectra import SpectraSmootherGaussian
    >>> from resistics.testing import spectra_data_basic
    >>> spec_data = spectra_data_basic()
    >>> smooth_data = SpectraSmootherGaussian().run(spec_data)

    Look at the results for the two windows

    >>> spec_data.data[0][0,0]
    array([0.+0.j, 1.+1.j, 2.+2.j, 3.+3.j, 4.+4.j, 5.+5.j, 6.+6.j, 7.+7.j,
           8.+8.j, 9.+9.j])
    >>> smooth_data.data[0][0,0]
    array([0.42704095+0.42704095j, 1.06795587+1.06795587j,
           2.00483335+2.00483335j, 3.00013383+3.00013383j,
           4.        +4.j        , 5.        +5.j        ,
           5.99986617+5.99986617j, 6.99516665+6.99516665j,
           7.93204413+7.93204413j, 8.57295905+8.57295905j])
    """

    sigma: float = 3

    def run(self, spec_data: SpectraData) -> SpectraData:
        """
        Run Gaussian filtering of spectra data

        Parameters
        ----------
        spec_data : SpectraData
            Input spectra data

        Returns
        -------
        SpectraData
            Output spectra data
        """
        import scipy.ndimage as ndimage

        data = {}
        logger.info(f"Smoothing frequencies with gaussian filter, sigma {self.sigma}")
        messages = [f"Smoothing frequencies with gaussian filter, sigma {self.sigma}"]
        for ilevel in range(spec_data.metadata.n_levels):
            data[ilevel] = ndimage.gaussian_filter1d(
                spec_data.get_level(ilevel), 1, axis=-1
            )
            messages.append(f"Smoothed level {ilevel} with gaussian filter")
        metadata = SpectraMetadata(**spec_data.metadata.dict())
        metadata.history.add_record(self._get_record(messages))
        logger.info("Fourier coefficients calculated at evaluation frequencies")
        return SpectraData(metadata, data)



"""
Module defining transfer functions
"""
from typing import List, Optional, Dict, Any, Union
from pydantic import validator, constr
import numpy as np
import plotly.graph_objects as go

from resistics.common import Metadata


class Component(Metadata):
    """
    Data class for a single component in a Transfer function

    Example
    -------
    >>> from resistics.transfunc import Component
    >>> component = Component(real=[1, 2, 3, 4, 5], imag=[-5, -4, -3, -2 , -1])
    >>> component.get_value(0)
    (1-5j)
    >>> component.to_numpy()
    array([1.-5.j, 2.-4.j, 3.-3.j, 4.-2.j, 5.-1.j])
    """

    real: List[float]
    """The real part of the component"""
    imag: List[float]
    """The complex part of the component"""

    def get_value(self, eval_idx: int) -> complex:
        """Get the value for an evaluation frequency"""
        return self.real[eval_idx] + 1j * self.imag[eval_idx]

    def to_numpy(self) -> np.ndarray:
        """Get the component as a numpy complex array"""
        return np.array(self.real) + 1j * np.array(self.imag)


def get_component_key(out_chan: str, in_chan: str) -> str:
    """
    Get key for out channel and in channel combination in the solution

    Parameters
    ----------
    out_chan : str
        The output channel
    in_chan : str
        The input channel

    Returns
    -------
    str
        The component key

    Examples
    --------
    >>> from resistics.regression import get_component_key
    >>> get_component_key("Ex", "Hy")
    'ExHy'
    """
    return f"{out_chan}{in_chan}"


class TransferFunction(Metadata):
    """
    Define a generic transfer function

    This class is a describes generic transfer function, including:

    - The output channels for the transfer function
    - The input channels for the transfer function
    - The cross channels for the transfer function

    The cross channels are the channels that will be used to calculate out the
    cross powers for the regression.

    This generic parent class has no implemented plotting function. However,
    child classes may have a plotting function as different transfer functions
    may need different types of plots.

    .. note::

        Users interested in writing a custom transfer function should inherit
        from this generic Transfer function

    See Also
    --------
    ImpandanceTensor : Transfer function for the MT impedance tensor
    Tipper : Transfer function for the MT tipper

    Examples
    --------
    A generic example

    >>> tf = TransferFunction(variation="example", out_chans=["bye", "see you", "ciao"], in_chans=["hello", "hi_there"])
    >>> print(tf.to_string())
    | bye      |   | bye_hello         bye_hi_there      | | hello    |
    | see you  | = | see you_hello     see you_hi_there  | | hi_there |
    | ciao     |   | ciao_hello        ciao_hi_there     |

    Combining the impedance tensor and the tipper into one TransferFunction

    >>> tf = TransferFunction(variation="combined", out_chans=["Ex", "Ey"], in_chans=["Hx", "Hy", "Hz"])
    >>> print(tf.to_string())
    | Ex |   | Ex_Hx Ex_Hy Ex_Hz | | Hx |
    | Ey | = | Ey_Hx Ey_Hy Ey_Hz | | Hy |
                                   | Hz |
    """

    _types: Dict[str, type] = {}
    """Store types which will help automatic instantiation"""
    name: Optional[str] = None
    """The name of the transfer function, this will be set automatically"""
    variation: constr(max_length=16) = "generic"
    """A short additional bit of information about this variation"""
    out_chans: List[str]
    """The output channels"""
    in_chans: List[str]
    """The input channels"""
    cross_chans: Optional[List[str]] = None
    """The channels to use for calculating the cross spectra"""
    n_out: Optional[int] = None
    """The number of output channels"""
    n_in: Optional[int] = None
    """The number of input channels"""
    n_cross: Optional[int] = None
    """The number of cross power channels"""

    def __init_subclass__(cls) -> None:
        """
        Used to automatically register child transfer functions in `_types`

        When a TransferFunction child class is imported, it is added to the base
        TransferFunction _types variable. Later, this dictionary of class types
        can be used to initialise a specific child transfer function from a
        dictonary as long as that specific child transfer fuction has already
        been imported and it is called from a pydantic class that will validate
        the inputs.

        The intention of this method is to support initialising transfer
        functions from JSON files. This is a similar approach to
        ResisticsProcess.
        """
        cls._types[cls.__name__] = cls

    @classmethod
    def __get_validators__(cls):
        """Get the validators that will be used by pydantic"""
        yield cls.validate

    @classmethod
    def validate(
        cls, value: Union["TransferFunction", Dict[str, Any]]
    ) -> "TransferFunction":
        """
        Validate a TransferFunction

        Parameters
        ----------
        value : Union[TransferFunction, Dict[str, Any]]
            A TransferFunction child class or a dictionary

        Returns
        -------
        TransferFunction
            A TransferFunction or TransferFunction child class

        Raises
        ------
        ValueError
            If the value is neither a TransferFunction or a dictionary
        KeyError
            If name is not in the dictionary
        ValueError
            If initialising from dictionary fails

        Examples
        --------
        The following example will show how a child TransferFunction class
        can be instantiated using a dictionary and the parent TransferFunction
        (but only as long as that child class has been imported).

        >>> from resistics.transfunc import TransferFunction

        Show known TransferFunction types in built into resistics

        >>> for entry in TransferFunction._types.items():
        ...     print(entry)
        ('ImpedanceTensor', <class 'resistics.transfunc.ImpedanceTensor'>)
        ('Tipper', <class 'resistics.transfunc.Tipper'>)

        Now let's initialise an ImpedanceTensor from the base TransferFunction
        and a dictionary.

        >>> mytf = {"name": "ImpedanceTensor", "variation": "ecross", "cross_chans": ["Ex", "Ey"]}
        >>> test = TransferFunction(**mytf)
        Traceback (most recent call last):
        ...
        KeyError: 'out_chans'

        This is not quite what we were expecting. The generic TransferFunction
        requires out_chans to be defined, but they are not in the dictionary as
        the ImpedanceTensor child class defaults these. To get this to work,
        instead use the validate class method. This is the class method used by
        pydantic when instantiating.

        >>> mytf = {"name": "ImpedanceTensor", "variation": "ecross", "cross_chans": ["Ex", "Ey"]}
        >>> test = TransferFunction.validate(mytf)
        >>> test.summary()
        {
            'name': 'ImpedanceTensor',
            'variation': 'ecross',
            'out_chans': ['Ex', 'Ey'],
            'in_chans': ['Hx', 'Hy'],
            'cross_chans': ['Ex', 'Ey'],
            'n_out': 2,
            'n_in': 2,
            'n_cross': 2
        }

        That's more like it. This will raise errors if an unknown type of
        TransferFunction is received.

        >>> mytf = {"name": "NewTF", "cross_chans": ["Ex", "Ey"]}
        >>> test = TransferFunction.validate(mytf)
        Traceback (most recent call last):
        ...
        ValueError: Unable to initialise NewTF from dictionary

        Or if the dictionary does not have a name key

        >>> mytf = {"cross_chans": ["Ex", "Ey"]}
        >>> test = TransferFunction.validate(mytf)
        Traceback (most recent call last):
        ...
        KeyError: 'No name provided for initialisation of TransferFunction'

        Unexpected inputs will also raise an error

        >>> test = TransferFunction.validate(5)
        Traceback (most recent call last):
        ...
        ValueError: TransferFunction unable to initialise from <class 'int'>
        """
        if isinstance(value, TransferFunction):
            return value
        if not isinstance(value, dict):
            raise ValueError(
                f"TransferFunction unable to initialise from {type(value)}"
            )
        if "name" not in value:
            raise KeyError("No name provided for initialisation of TransferFunction")
        # check if it is a TransferFunction
        name = value.pop("name")
        if name == "TransferFunction":
            return cls(**value)
        # check other known Transfer Functions
        try:
            return cls._types[name](**value)
        except Exception:
            raise ValueError(f"Unable to initialise {name} from dictionary")

    @validator("name", always=True)
    def validate_name(cls, value: Union[str, None]) -> str:
        """Inialise the name attribute of the transfer function"""
        if value is None:
            return cls.__name__
        return value

    @validator("cross_chans", always=True)
    def validate_cross_chans(
        cls, value: Union[None, List[str]], values: Dict[str, Any]
    ) -> List[str]:
        """Validate cross spectra channels"""
        if value is None:
            return values["in_chans"]
        return value

    @validator("n_out", always=True)
    def validate_n_out(cls, value: Union[None, int], values: Dict[str, Any]) -> int:
        """Validate number of output channels"""
        if value is None:
            return len(values["out_chans"])
        return value

    @validator("n_in", always=True)
    def validate_n_in(cls, value: Union[None, int], values: Dict[str, Any]) -> int:
        """Validate number of input channels"""
        if value is None:
            return len(values["in_chans"])
        return value

    @validator("n_cross", always=True)
    def validate_n_cross(cls, value: Union[None, int], values: Dict[str, Any]) -> int:
        """Validate number of cross channels"""
        if value is None:
            return len(values["cross_chans"])
        return value

    def n_eqns_per_output(self) -> int:
        """Get the number of equations per output"""
        return len(self.cross_chans)

    def n_regressors(self) -> int:
        """Get the number of regressors"""
        return self.n_in

    def to_string(self):
        """Get the transfer function as as string"""
        n_lines = max(len(self.in_chans), len(self.out_chans))
        lens = [len(x) for x in self.in_chans] + [len(x) for x in self.out_chans]
        max_len = max(lens)
        line_equals = (n_lines - 1) // 2
        outstr = ""
        for il in range(n_lines):
            out_chan = self._out_chan_string(il, max_len)
            in_chan = self._in_chan_string(il, max_len)
            tensor = self._tensor_string(il, max_len)
            eq = "=" if il == line_equals else " "
            outstr += f"{out_chan} {eq} {tensor} {in_chan}\n"
        return outstr.rstrip("\n")

    def _out_chan_string(self, il: int, max_len: int) -> str:
        """Get the out channels string"""
        if il >= self.n_out:
            empty_len = max_len + 4
            return f"{'':{empty_len}s}"
        return f"| { self.out_chans[il]:{max_len}s} |"

    def _in_chan_string(self, il: int, max_len: int) -> str:
        """Get the in channel string"""
        if il >= self.n_in:
            return ""
        return f"| { self.in_chans[il]:{max_len}s} |"

    def _tensor_string(self, il: int, max_len: int) -> str:
        """Get the tensor string"""
        if il >= self.n_out:
            element_len = ((max_len * 2 + 1) + 1) * self.n_in + 3
            return f"{'':{element_len}s}"
        elements = "| "
        for chan in self.in_chans:
            component = f"{self.out_chans[il]}_{chan}"
            elements += f"{component:{2*max_len + 1}s} "
        elements += "|"
        return elements


class ImpedanceTensor(TransferFunction):
    """
    Standard magnetotelluric impedance tensor

    Notes
    -----
    Information about data units

    - Magnetic permeability in nT . m / A
    - Electric (E) data is in mV/m
    - Magnetic (H) data is in nT
    - Z = E/H is in mV / m . nT
    - Units of resistance = Ohm = V / A

    Examples
    --------
    >>> from resistics.transfunc import ImpedanceTensor
    >>> tf = ImpedanceTensor()
    >>> print(tf.to_string())
    | Ex | = | Ex_Hx Ex_Hy | | Hx |
    | Ey |   | Ey_Hx Ey_Hy | | Hy |
    """

    variation: constr(max_length=16) = "default"
    out_chans: List[str] = ["Ex", "Ey"]
    in_chans: List[str] = ["Hx", "Hy"]

    def get_resistivity(self, periods: np.ndarray, component: Component) -> np.ndarray:
        """Get the apparent resistivity for a component"""
        squared = np.power(np.absolute(component.to_numpy()), 2)
        return 0.2 * periods * squared

    def get_phase(self, key: str, component: Component) -> np.ndarray:
        """Get the phase for a component"""
        phase = np.angle(component.to_numpy())
        # unwrap into specific quadrant and convert to degrees
        phase = np.unwrap(phase) * 180 / np.pi
        if key == "ExHx" or key == "ExHy":
            phase = np.mod(phase, 360) - 180
        return phase

    def plot(
        self,
        freqs: List[float],
        components: Dict[str, Component],
        to_plot: Optional[List[str]] = None,
        x_lim: Optional[List[float]] = None,
        res_lim: Optional[List[float]] = None,
        phs_lim: Optional[List[float]] = None,
    ) -> go.Figure:
        """
        Plot the impedance tensor

        Parameters
        ----------
        freqs : List[float]
            The x axis frequencies
        components : Dict[str, Component]
            The component data
        to_plot : Optional[List[str]], optional
            The components of the impedance tensor to plot, by default None,
            which will plot all of them
        x_lim : Optional[List[float]], optional
            The x limits, to be provided as powers of 10, by default None. For
            example, for 0.001, use -3
        res_lim : Optional[List[float]], optional
            The y limits for resistivity, to be provided as powers of 10, by
            default None. For example, for 1000, use 3
        phs_lim : Optional[List[float]], optional
            The phase limits, by default None

        Returns
        -------
        go.Figure
            Plotly figure
        """
        from plotly.subplots import make_subplots

        periods = np.reciprocal(freqs)
        if to_plot is None:
            to_plot = ["ExHy", "EyHx", "ExHx", "EyHy"]
        if x_lim is None:
            x_lim = [-3, 5]
        if res_lim is None:
            res_lim = [-2, 6]
        if phs_lim is None:
            phs_lim = [-10, 100]

        colors = {"ExHx": "orange", "EyHy": "green", "ExHy": "red", "EyHx": "blue"}
        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.08,
            subplot_titles=["Apparent resistivity", "Phase"],
        )
        fig.update_layout(width=1000, autosize=True)
        # x axes
        fig.update_xaxes(title_text="Period (s)", type="log", range=x_lim, row=1, col=1)
        fig.update_xaxes(showticklabels=True, row=1, col=1)
        fig.update_xaxes(title_text="Period (s)", type="log", range=x_lim, row=2, col=1)
        fig.update_xaxes(showticklabels=True, row=2, col=1)
        # y axes
        fig.update_yaxes(title_text="App. resistivity (Ohm m)", row=1, col=1)
        fig.update_yaxes(type="log", range=res_lim, row=1, col=1)
        fig.update_yaxes(scaleanchor="x", scaleratio=1, row=1, col=1)
        fig.update_yaxes(title_text="Phase (degrees)", range=phs_lim, row=2, col=1)
        for comp in to_plot:
            res = self.get_resistivity(periods, components[comp])
            phs = self.get_phase(comp, components[comp])
            scatter = go.Scatter(
                x=periods,
                y=res,
                mode="lines+markers",
                marker=dict(color=colors[comp]),
                line=dict(color=colors[comp]),
                legendgroup=comp,
                name=comp,
            )
            fig.add_trace(scatter, row=1, col=1)
            scatter = go.Scatter(
                x=periods,
                y=phs,
                mode="lines+markers",
                marker=dict(color=colors[comp]),
                line=dict(color=colors[comp]),
                legendgroup=comp,
                name=comp,
                showlegend=False,
            )
            fig.add_trace(scatter, row=2, col=1)
        return fig


class Tipper(TransferFunction):
    """
    Magnetotelluric tipper

    The tipper components are Tx = HzHx and Ty = HzHy

    The tipper length is sqrt(Re(Tx)^2 + Re(Ty)^2)

    The tipper angle is arctan (Re(Ty)/Re(Tx))

    Notes
    -----
    Information about units

    - Tipper T = H/H is dimensionless

    Examples
    --------
    >>> from resistics.transfunc import Tipper
    >>> tf = Tipper()
    >>> print(tf.to_string())
    | Hz | = | Hz_Hx Hz_Hy | | Hx |
                             | Hy |
    """

    variation: constr(max_length=16) = "default"
    out_chans: List[str] = ["Hz"]
    in_chans: List[str] = ["Hx", "Hy"]

    def get_length(self, components: Dict[str, Component]) -> np.ndarray:
        """Get the tipper length"""
        txRe = components["HzHx"].real
        tyRe = components["HzHy"].real
        return np.sqrt(np.power(txRe, 2) + np.power(tyRe, 2))

    def get_real_angle(self, components: Dict[str, Component]) -> np.ndarray:
        """Get the real angle"""
        txRe = np.array(components["HzHx"].real)
        tyRe = np.array(components["HzHy"].real)
        return np.arctan(tyRe / txRe) * 180 / np.pi

    def get_imag_angle(self, components: Dict[str, Component]) -> np.ndarray:
        """Get the imaginary angle"""
        txIm = np.array(components["HzHx"].imag)
        tyIm = np.array(components["HzHy"].imag)
        return np.arctan(tyIm / txIm) * 180 / np.pi

    def plot(
        self,
        freqs: List[float],
        components: Dict[str, Component],
        x_lim: Optional[List[float]] = None,
        len_lim: Optional[List[float]] = None,
        ang_lim: Optional[List[float]] = None,
    ) -> go.Figure:
        """
        Plot the impedance tensor

        .. warning::

            This probably needs further checking and verification

        Parameters
        ----------
        freqs : List[float]
            The x axis frequencies
        components : Dict[str, Component]
            The component data
        x_lim : Optional[List[float]], optional
            The x limits, to be provided as powers of 10, by default None. For
            example, for 0.001, use -3
        len_lim : Optional[List[float]], optional
            The y limits for tipper length, to be provided as powers of 10, by
            default None. For example, for 1000, use 3
        ang_lim : Optional[List[float]], optional
            The angle limits, by default None

        Returns
        -------
        go.Figure
            Plotly figure
        """
        import warnings
        from plotly.subplots import make_subplots

        warnings.warn("Plotting of tippers needs further verification")

        periods = np.reciprocal(freqs)
        if x_lim is None:
            x_lim = [-3, 5]
        if len_lim is None:
            len_lim = [-2, 6]
        if ang_lim is None:
            ang_lim = [-10, 100]

        fig = make_subplots(
            rows=2,
            cols=1,
            shared_xaxes=True,
            vertical_spacing=0.08,
            subplot_titles=["Length", "Angles"],
        )
        fig.update_layout(width=1000, autosize=True)
        # x axes
        fig.update_xaxes(title_text="Period (s)", type="log", range=x_lim, row=1, col=1)
        fig.update_xaxes(showticklabels=True, row=1, col=1)
        fig.update_xaxes(title_text="Period (s)", type="log", range=x_lim, row=2, col=1)
        fig.update_xaxes(showticklabels=True, row=2, col=1)
        # y axes
        fig.update_yaxes(title_text="Tipper length", row=1, col=1)
        # fig.update_yaxes(type="log", row=1, col=1)
        # fig.update_yaxes(scaleanchor="x", scaleratio=1, row=1, col=1)
        fig.update_yaxes(title_text="Angle (degrees)", row=2, col=1)
        # plot the tipper length
        scatter = go.Scatter(
            x=periods,
            y=self.get_length(components),
            mode="lines+markers",
            marker=dict(color="red"),
            line=dict(color="red"),
            name="Tipper length",
        )
        fig.add_trace(scatter, row=1, col=1)
        # plot the real angle
        scatter = go.Scatter(
            x=periods,
            y=self.get_real_angle(components),
            mode="lines+markers",
            marker=dict(color="green"),
            line=dict(color="green"),
            name="Real angle",
        )
        fig.add_trace(scatter, row=2, col=1)
        # plot the imag angle
        scatter = go.Scatter(
            x=periods,
            y=self.get_imag_angle(components),
            mode="lines+markers",
            marker=dict(color="blue"),
            line=dict(color="blue"),
            name="Imag angle",
        )
        fig.add_trace(scatter, row=2, col=1)
        return fig



"""
This module is the main interface to resistics and includes:

- Classes and functions for making, loading and using resistics projects
- Functions for processing data
"""
from loguru import logger
from typing import Optional, Dict, Union, Any
from pathlib import Path
from datetime import datetime

from resistics.errors import MetadataReadError, TimeDataReadError
from resistics.common import ResisticsProcess, ResisticsModel
from resistics.config import Configuration, get_default_configuration
from resistics.project import PROJ_FILE, PROJ_DIRS, ProjectMetadata, Project
from resistics.project import Site, Measurement
from resistics.sampling import DateTimeLike, HighResDateTime
from resistics.time import TimeData
from resistics.decimate import DecimationParameters
from resistics.decimate import DecimatedData
from resistics.window import WindowedData
from resistics.spectra import SpectraData
from resistics.gather import GatheredData
from resistics.regression import RegressionInputData, Solution


class ProjectCreator(ResisticsProcess):
    """Process to create a project"""

    dir_path: Path
    metadata: ProjectMetadata

    def run(self):
        """
        Create the project

        Raises
        ------
        ProjectCreateError
            If an existing project found
        """
        from resistics.errors import ProjectCreateError
        from resistics.common import is_dir, assert_dir

        metadata_path = self.dir_path / PROJ_FILE
        if self.dir_path.exists() and is_dir(self.dir_path) and metadata_path.exists():
            raise ProjectCreateError(
                self.dir_path, "Existing project found, try loading"
            )
        elif self.dir_path.exists():
            logger.warning("Directory already exists, the project will be saved here")
            assert_dir(self.dir_path)
        else:
            logger.info("Making directory for project")
            self.dir_path.mkdir(parents=True)
        self._make_subdirs()
        self.metadata.write(metadata_path)
        logger.info(f"Project created in {self.dir_path}")

    def _make_subdirs(self):
        """Make project subdirectories"""
        for data_type, subdir in PROJ_DIRS.items():
            subdir_path = self.dir_path / subdir
            if not subdir_path.exists():
                logger.info(f"Making {data_type} data subdirectory: {subdir_path}")
                subdir_path.mkdir()


def new(dir_path: Union[Path, str], proj_info: Dict[str, Any]) -> bool:
    """
    Create a new project

    Parameters
    ----------
    dir_path : Union[Path, str]
        The directory to create the project in
    proj_info : Dict[str, Any]
        Any project details

    Returns
    -------
    bool
        True if the creator was successful
    """
    if isinstance(dir_path, str):
        dir_path = Path(dir_path)
    metadata = ProjectMetadata(**proj_info)
    ProjectCreator(dir_path=dir_path, metadata=metadata).run()
    return True


class ProjectLoader(ResisticsProcess):
    """Project loader"""

    dir_path: Path

    def run(self, config: Configuration) -> Project:
        """
        Load a project

        Parameters
        ----------
        config : Configuration
            The configuration for the purposes of getting the time readers

        Returns
        -------
        Project
            Project instance

        Raises
        ------
        ProjectLoadError
            If the resistcs project metadata is not found
        """
        from resistics.errors import ProjectLoadError
        from resistics.common import assert_dir, dir_subdirs

        assert_dir(self.dir_path)
        metadata_path = self.dir_path / PROJ_FILE
        if not metadata_path.exists():
            raise ProjectLoadError(
                self.dir_path, f"Resistics project file {metadata_path} not found"
            )
        self._check_subdirs()

        metadata = ProjectMetadata.parse_file(metadata_path)
        time_subdirs = dir_subdirs(self.dir_path / PROJ_DIRS["time"])
        sites = {}
        for site_dir in time_subdirs:
            site = self._load_site(site_dir, config)
            sites[site_dir.name] = site
        if len(sites) > 0:
            begin_time = min([x.begin_time for x in sites.values()])
            end_time = max([x.end_time for x in sites.values()])
        else:
            begin_time = datetime.now()
            end_time = datetime.now()
        return Project(
            dir_path=self.dir_path,
            metadata=metadata,
            begin_time=begin_time,
            end_time=end_time,
            sites=sites,
        )

    def _check_subdirs(self) -> bool:
        """Returns True if all require project subdirectories exist otherwise False"""
        from resistics.common import assert_dir

        for subdir in PROJ_DIRS.values():
            subdir_path = self.dir_path / subdir
            assert_dir(subdir_path)
            return False
        return True

    def _load_site(self, site_dir: Path, config: Configuration) -> Site:
        """Load a Site"""
        from resistics.common import dir_subdirs

        subdirs = dir_subdirs(site_dir)
        measurements = {}
        for meas_dir in subdirs:
            meas = self._load_measurement(site_dir.name, meas_dir, config)
            if meas is not None:
                measurements[meas_dir.name] = meas
        if len(measurements) > 0:
            begin_time = min([x.metadata.first_time for x in measurements.values()])
            end_time = max([x.metadata.last_time for x in measurements.values()])
        else:
            begin_time = datetime.now()
            end_time = datetime.now()
        return Site(
            dir_path=site_dir,
            begin_time=begin_time,
            end_time=end_time,
            measurements=measurements,
        )

    def _load_measurement(
        self, site_name: str, meas_dir: Path, config: Configuration
    ) -> Union[Measurement, None]:
        """
        Load a measurement

        The loader tries to use any TimeReader provided in the configuration to
        load the measurement. If no compatible reader is found, the measurement
        will be ignored.

        Parameters
        ----------
        site_name : str
            The name of the Site
        meas_dir : Path
            The measurement subdirectory in the site time directory
        config : Configuration
            Configuration which is used for the time readers

        Returns
        -------
        Union[Measurement, None]
            Measurement if reading was successful, else None
        """
        for reader in config.time_readers:
            try:
                metadata = reader.run(dir_path=meas_dir, metadata_only=True)
                logger.info(f"Read measurement {meas_dir} with {reader.name}")
                return Measurement(
                    site_name=site_name,
                    dir_path=meas_dir,
                    metadata=metadata,
                    reader=reader,
                )
            except Exception:
                logger.debug(
                    f"Failed to read measurement {meas_dir} with {reader.name}"
                )
        logger.error(f"No reader found for measurement {meas_dir}")
        return None


class ResisticsEnvironment(ResisticsModel):
    """
    A Resistics environment which combines a project and a configuration
    """

    proj: Project
    """The project"""
    config: Configuration
    """The configuration for processing"""


def load(
    dir_path: Union[Path, str], config: Optional[Configuration] = None
) -> ResisticsEnvironment:
    """
    Load an existing project into a ResisticsEnvironment

    Parameters
    ----------
    dir_path : Union[Path, str]
        The project directory
    config : Optional[Configuration], optional
        A configuration of parameters to use

    Returns
    -------
    ResisticsEnvironment
        The ResisticsEnvironment combining a project and a configuration

    Raises
    ------
    ProjectLoadError
        If the loading failed
    """
    if config is None:
        config = get_default_configuration()
    if isinstance(dir_path, str):
        dir_path = Path(dir_path)
    proj = ProjectLoader(dir_path=dir_path).run(config)
    return ResisticsEnvironment(proj=proj, config=config)


def reload(resenv: ResisticsEnvironment) -> ResisticsEnvironment:
    """
    Reload the project in the ResisticsEnvironment

    Parameters
    ----------
    resenv : ResisticsEnvironment
        The current resistics environment

    Returns
    -------
    ResisticsEnvironment
        The resistics environment with the project reloaded
    """
    return load(dir_path=resenv.proj.dir_path, config=resenv.config)


def run_time_processors(config: Configuration, time_data: TimeData) -> TimeData:
    """
    Process time data

    Parameters
    ----------
    config : Configuration
        The configuration
    time_data : TimeData
        Time data to process

    Returns
    -------
    TimeData
        Process time data
    """
    for process in config.time_processors:
        logger.info(f"Running processor {process.name}")
        time_data = process.run(time_data)
    return time_data


def run_decimation(
    config: Configuration,
    time_data: TimeData,
    dec_params: Optional[DecimationParameters] = None,
) -> DecimatedData:
    """
    Decimate TimeData

    Parameters
    ----------
    config : Configuration
        The configuration
    time_data : TimeData
        Time data to decimate
    dec_params : DecimationParameters
        Number of levels, decimation factors etc.

    Returns
    -------
    DecimatedData
        Decimated time data
    """
    logger.info("Decimating time data")
    if dec_params is None:
        dec_params = config.dec_setup.run(time_data.metadata.fs)
    return config.decimator.run(dec_params, time_data)


def run_windowing(
    config: Configuration, ref_time: HighResDateTime, dec_data: DecimatedData
) -> WindowedData:
    """
    Window time data

    Parameters
    ----------
    config : Configuration
        The configuration
    ref_time : HighResDateTime
        The reference time
    dec_data : DecimatedData
        Decimated data to window

    Returns
    -------
    WindowedData
        The windowed data
    """
    logger.info("Windowing time data")
    win_params = config.win_setup.run(dec_data.metadata.n_levels, dec_data.metadata.fs)
    return config.windower.run(ref_time, win_params, dec_data)


def run_fft(config: Configuration, win_data: WindowedData) -> SpectraData:
    """
    Run Fourier transform

    Parameters
    ----------
    config : Configuration
        The configuration
    win_data : WindowedData
        Windowed data

    Returns
    -------
    SpectraData
        Fourier transformed windowed data
    """
    logger.info("Calculating spectra data")
    return config.fourier.run(win_data)


def run_spectra_processors(
    config: Configuration, spec_data: SpectraData
) -> SpectraData:
    """
    Run any spectra processors

    Parameters
    ----------
    config : Configuration
        The configuration
    spec_data : SpectraData
        Spectra data

    Returns
    -------
    SpectraData
        Processed spectra data
    """
    for process in config.spectra_processors:
        logger.info(f"Running processor {process.name}")
        spec_data = process.run(spec_data)
    return spec_data


def run_evals(
    config: Configuration, dec_params: DecimationParameters, spec_data: SpectraData
) -> SpectraData:
    """
    Run evaluation frequency data calculator

    Parameters
    ----------
    config : Configuration
        The configuration
    dec_params : DecimationParameters
        Decimation parameters with the evaluation frequencies
    spec_data : SpectraData
        The spectra data

    Returns
    -------
    SpectraData
        Spectra data at evaluation frequencies
    """
    logger.info("Calculating fourier coefficients at evaluation frequencies")
    return config.evals.run(dec_params, spec_data)


def run_sensor_calibration(
    config: Configuration, calibration_path: Path, spec_data: SpectraData
) -> SpectraData:
    """
    Run calibration

    Parameters
    ----------
    config : Configuration
        The configuration
    calibration_path : Path
        Path to calibration data
    spec_data : SpectraData
        Spectra data to calibrate

    Returns
    -------
    SpectraData
        Calibrated spectra data
    """
    logger.info("Calibrating time data")
    return config.sensor_calibrator.run(calibration_path, spec_data)


def run_regression_preparer(
    config: Configuration, gathered_data: GatheredData
) -> RegressionInputData:
    """
    Prepare linear regression data

    Parameters
    ----------
    config : Configuration
        The configuration
    gathered_data : GatheredData
        Gathered data to input into the regression

    Returns
    -------
    RegressionInputData
        Regression inputs for all evaluation frequencies
    """
    logger.info("Preparing regression input data")
    return config.regression_preparer.run(config.tf, gathered_data)


def run_solver(config: Configuration, reg_data: RegressionInputData) -> Solution:
    """
    Run the regression solver

    Parameters
    ----------
    config : Configuration
        The configuration
    reg_data : RegressionInputData
        The regression input data

    Returns
    -------
    Solution
        Transfer function estimate
    """
    logger.info(f"Running solver {config.solver.name}")
    return config.solver.run(reg_data)


def quick_read(config: Configuration, dir_path: Path) -> TimeData:
    """
    Read time data folder

    Parameters
    ----------
    config : Configuration
        Configuration with appropriate readers, otherwise returns None
    dir_path : Path
        The directory path to read

    Returns
    -------
    TimeData
        The read TimeData

    Raises
    ------
    TimeDataReadError
        If unable to read data
    """
    for reader in config.time_readers:
        logger.info(f"Attempting to read data with reader {reader.name}")
        try:
            return reader.run(dir_path)
        except MetadataReadError:
            logger.debug(f"Unable to read metadata with reader {reader.name}")
        except TimeDataReadError:
            logger.debug(f"Failed reading time data with reader {reader.name}")
        except Exception:
            logger.debug("Unknown problem reading time data")
    reader_names = [reader.name for reader in config.time_readers]
    raise TimeDataReadError(
        dir_path, f"Unable to read time data with readers {reader_names}"
    )


def quick_view(
    dir_path: Path,
    config: Optional[Configuration] = None,
    decimate: bool = False,
    max_pts: int = 10_000,
):
    """
    Quick plotting of time data

    Parameters
    ----------
    dir_path : Path
        The directory path
    config : Optional[Configuration], optional
        The configuration with the required time readers, by default None
    decimate : bool, optional
        Boolean flag for decimating, by default False
    max_pts : Optional[int], optional
        Max points in lttb decimation, by default 10_000

    Raises
    ------
    ValueError
        If time data fails reading
    """
    logger.info(f"Plotting data in {dir_path}")
    if config is None:
        config = get_default_configuration()

    time_data = quick_read(config, dir_path)
    time_data = run_time_processors(config, time_data)
    if not decimate:
        time_data.plot(max_pts=max_pts).show()
        return
    dec_params = config.dec_setup.run(time_data.metadata.fs)
    dec_data = run_decimation(config, time_data, dec_params=dec_params)
    dec_data.plot(max_pts=max_pts).show()


def quick_tf(
    dir_path: Path,
    config: Optional[Configuration] = None,
    calibration_path: Optional[Path] = None,
) -> Solution:
    """
    Quickly calculate out a transfer function for time data in its own directory

    Parameters
    ----------
    dir_path : Path
        The directory path
    config : Optional[Configuration], optional
        A configuration instance, by default None
    calibration_path : Optional[Path], optional
        The path to the calibration data, by default None

    Returns
    -------
    Solution
        Transfer function estimate
    """
    from resistics.gather import QuickGather

    logger.info(f"Processing data in {dir_path}")
    if config is None:
        config = get_default_configuration()

    time_data = quick_read(config, dir_path)
    ref_time = time_data.metadata.first_time
    time_data = run_time_processors(config, time_data)
    dec_params = config.dec_setup.run(time_data.metadata.fs)
    dec_data = run_decimation(config, time_data, dec_params=dec_params)
    win_data = run_windowing(config, ref_time, dec_data)
    spec_data = run_fft(config, win_data)
    spec_data = run_spectra_processors(config, spec_data)
    eval_data = run_evals(config, dec_params, spec_data)
    if calibration_path is not None:
        eval_data = run_sensor_calibration(config, calibration_path, eval_data)
    gathered_data = QuickGather().run(dir_path, dec_params, config.tf, eval_data)
    reg_data = run_regression_preparer(config, gathered_data)
    return run_solver(config, reg_data)


def process_time(
    resenv: ResisticsEnvironment,
    site_name: str,
    meas_name: str,
    out_site: str,
    out_meas: str,
    from_time: Optional[DateTimeLike] = None,
    to_time: Optional[DateTimeLike] = None,
) -> None:
    """
    Process time data and save as a new measurement

    This is useful when resampling data to use with other measurements

    Parameters
    ----------
    resenv : ResisticsEnvironment
        The resistics environment
    site_name : str
        The name of the site with the data to process
    meas_name : str
        The name of the measurement to process
    out_site : str
        The site to output the data to
    out_meas : str
        The name of the measurement to output the data to
    from_time : Optional[DateTimeLike], optional
        Time to get the time data from, by default None
    to_time : Optional[DateTimeLike], optional
        Time to get the time data up to, by default None
    """
    from resistics.time import TimeWriterNumpy
    from resistics.project import get_meas_time_path

    logger.info(f"Running time processors on meas {meas_name} from site {site_name}")
    meas = resenv.proj[site_name][meas_name]
    time_data = meas.reader.run(
        meas.dir_path, metadata=meas.metadata, from_time=from_time, to_time=to_time
    )
    time_data = run_time_processors(resenv.config, time_data)
    out_path = get_meas_time_path(resenv.proj.dir_path, out_site, out_meas)
    writer = TimeWriterNumpy()
    writer.run(out_path, time_data)


def process_time_to_evals(
    resenv: ResisticsEnvironment, site_name: str, meas_name: str
) -> None:
    """
    Process from time data to Fourier spectra

    Parameters
    ----------
    resenv : ResisticsEnvironment
        The resistics environment containing the project and configuration
    site_name : str
        The name of the site
    meas_name : str
        The name of the measurement to process
    """
    from resistics.project import get_calibration_path, get_meas_evals_path
    from resistics.spectra import SpectraDataWriter

    proj = resenv.proj
    config = resenv.config
    calibration_path = get_calibration_path(proj.dir_path)

    logger.info(f"Processing measurement {site_name}, {meas_name}")
    meas = proj[site_name][meas_name]
    time_data = meas.reader.run(meas.dir_path, metadata=meas.metadata)
    time_data = run_time_processors(config, time_data)
    dec_params = config.dec_setup.run(time_data.metadata.fs)
    dec_data = run_decimation(config, time_data, dec_params=dec_params)
    win_data = run_windowing(config, proj.metadata.ref_time, dec_data)
    spec_data = run_fft(config, win_data)
    spec_data = run_spectra_processors(config, spec_data)
    eval_data = run_evals(config, dec_params, spec_data)
    eval_data = run_sensor_calibration(config, calibration_path, eval_data)
    evals_path = get_meas_evals_path(
        proj.dir_path, meas.site_name, meas.name, config.name
    )
    logger.info(f"Saving evaluation frequency data to {evals_path}")
    SpectraDataWriter().run(evals_path, eval_data)


def process_evals_to_tf(
    resenv: ResisticsEnvironment,
    fs: float,
    out_site: str,
    in_site: Optional[str] = None,
    cross_site: Optional[str] = None,
    masks: Optional[Dict[str, str]] = None,
    postfix: Optional[str] = None,
) -> Solution:
    """
    Process spectra to transfer functions

    Parameters
    ----------
    resenv : ResisticsEnvironment
        The resistics environment
    fs : float
        The sampling frequency to process
    out_site : str
        The name of the output site
    in_site : Optional[str], optional
        The name of the input site, by default None. This should be used for
        intersite processing
    cross_site : Optional[str], optional
        The name of the cross site, by default None. This is usually the site to
        use as the remote reference.
    masks : Optional[Dict[str, str]], optional
        Any masks to apply, by default None
    postfix : Optional[str]
        String to add to the end of solution, by default None

    Returns
    -------
    Solution
        Transfer function estimate
    """
    from resistics.gather import Selector, ProjectGather
    from resistics.project import get_results_path, get_solution_name

    proj = resenv.proj
    config = resenv.config

    sites = [out_site]
    if in_site is not None:
        sites.append(in_site)
    if cross_site is not None:
        sites.append(cross_site)

    dec_params = config.dec_setup.run(fs)
    selection = Selector().run(config.name, proj, sites, dec_params)
    gathered_data = ProjectGather().run(
        config.name,
        proj,
        selection,
        config.tf,
        out_name=out_site,
        in_name=in_site,
        cross_name=cross_site,
    )
    reg_data = run_regression_preparer(config, gathered_data)
    solution = run_solver(config, reg_data)
    solution_path = get_results_path(proj.dir_path, out_site, config.name)
    if not solution_path.exists():
        solution_path.mkdir(parents=True)
    solution_name = get_solution_name(
        fs, solution.tf.name, solution.tf.variation, postfix
    )
    solution.write(solution_path / solution_name)
    return solution


def get_solution(
    resenv: ResisticsEnvironment,
    site_name: str,
    config_name: str,
    fs: float,
    tf_name: str,
    tf_var: str,
    postfix: Optional[str] = None,
) -> Solution:
    """
    Get a solution

    Parameters
    ----------
    resenv : ResisticsEnvironment
        The resistics environment
    site_name : str
        The site for which to get the solution
    config_name : str
        The configuration that was used
    fs : float
        The sampling frequency
    tf_name : str
        The transfer function name
    tf_var : str
        The transfer function variation
    postfix : Optional[str], optional
        Any postfix on the solution, by default None

    Returns
    -------
    Solution
        The solution
    """
    from resistics.project import get_results_path, get_solution_name

    proj = resenv.proj
    solution_path = get_results_path(proj.dir_path, site_name, config_name)
    solution_name = get_solution_name(fs, tf_name, tf_var, postfix)
    return Solution.parse_file(solution_path / solution_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['resistics']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'attotime==0.2.3',
 'loguru>=0.5.3,<0.6.0',
 'lttbc>=0.2.0,<0.3.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.2.3,<2.0.0',
 'plotly>=4.14.3,<5.0.0',
 'prettyprinter>=0.18.0,<0.19.0',
 'pydantic>=1.8.1,<2.0.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.6.2,<2.0.0',
 'tqdm>=4.61.0,<5.0.0']

setup_kwargs = {
    'name': 'resistics',
    'version': '1.0.0a3',
    'description': 'Python package for processing magnetotelluric data',
    'long_description': '## Welcome\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/resistics.svg)](https://pypi.org/project/resistics/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/resistics)](https://pypi.org/project/resistics/)\n[![Documentation Status](https://readthedocs.org/projects/resistics/badge/?version=latest)](https://resistics.readthedocs.io/en/latest/?badge=latest)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/resistics/resistics.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/resistics/resistics/context:python)\n[![codecov](https://codecov.io/gh/resistics/resistics/branch/master/graph/badge.svg?token=CXLJC9J7AW)](https://codecov.io/gh/resistics/resistics)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nResistics is a native Python 3.8+ package for the processing of magnetotelluric\n(MT) data. It incorporates robust processing methods and adopts a modular\napproach to processing which allows for customisation and future improvements\nto be quickly adopted.\n\n## Latest news\n\nResistics is moving to version 1.0.0 which will be a breaking change versus\nthe current stable version of 0.0.6. Currently, version 1.0.0 is available as a\npre-release on PYPI.\n\n- Documentation for 1.0.0: https://resistics.readthedocs.io/\n- Documentation for 0.0.6: https://resistics.io/\n\nWhen version 1.0.0 reaches a stable releases the documentation will move to the\nmain resistics.io site.\n\n## Audience\n\nResistics is intended for people who use magnetotelluric methods to estimate the\nsubsurface resistivity. This may be for furthering geological understanding, for\ngeothermal prospecting or for other purposes.\n\nThe package may have utility for the wider electromagnetic geophysics community.\n\n## Getting started\n\nTo install the pre-release of version 1.0.0\n\npython -m pip install resistics==1.0.0a0\n\nFor the stable 0.0.6 version\n\npython -m pip install resistics\n\n## Support and feature requests\n\nFeel free to submit issues, feature requests or ideas for improvements in the\nGithub issues section.\n',
    'author': 'Neeraj Shah',
    'author_email': 'resistics@outlook.com',
    'maintainer': 'Neeraj Shah',
    'maintainer_email': 'resistics@outlook.com',
    'url': 'https://www.resistics.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<3.10',
}


setup(**setup_kwargs)



"""
Module to help plotting various data
"""
from typing import List, Dict, Union, Tuple, Optional
import numpy as np
import plotly.graph_objects as go

from resistics.common import ResisticsData


def lttb_downsample(
    x: np.ndarray, y: np.ndarray, max_pts: int = 5000
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Downsample x, y for visualisation

    Parameters
    ----------
    x : np.ndarray
        x array
    y : np.ndarray
        y array
    max_pts : int, optional
        Maximum number of points after downsampling, by default 5000

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        (new_x, new_y), the downsampled x and y arrays

    Raises
    ------
    ValueError
        If the size of x does not match the size of y
    """
    import lttbc

    if x.size != y.size:
        raise ValueError(f"x size {x.size} must equal y size {y.size}")

    if max_pts >= x.size:
        return x, y

    x_dtype = x.dtype
    y_dtype = y.dtype
    nx, ny = lttbc.downsample(
        x.astype(np.float32),
        y.astype(np.float32),
        max_pts,
    )
    return nx.astype(x_dtype), ny.astype(y_dtype)


class PlotData1D(ResisticsData):
    """
    Class to help plot various 1-D data

    As this takes a full instance of x, it is not recommended for long time
    series as this will have a memory impact
    """

    def __init__(self, x: np.ndarray, data: np.ndarray, rows: List[str]):
        """
        Initialise

        Parameters
        ----------
        x : np.ndarray
            The x array
        data : np.ndarray
            The data, which is n_rows x n_x
        rows : List[str]
            The name of each row
        """
        self.x = x
        self.data = data
        self.rows = rows

    def __getitem__(self, row: str) -> np.ndarray:
        """Get data for a row by name"""
        index = self.rows.index(row)
        return self.data[index, :]

    def x_size(self) -> int:
        """Get the x size"""
        return len(self.x)

    def get_x(self, samples: Optional[np.ndarray] = None) -> np.ndarray:
        """Get x values"""
        if samples is None:
            return self.x
        return self.x[samples]


def figure_columns_as_lines(
    subplots: List[str],
    y_labels: Dict[str, str],
    x_label: str,
    title: Union[str, None] = None,
) -> go.Figure:
    """
    Get a figure for columnar data with specified subplots

    Parameters
    ----------
    subplots : List[str]
        The subplot titles
    y_labels : Dict[str, str]
        y labels for each subplot, with subplot as key and label as value
    x_label : str
        The x label, assumed to be the same for all subplots as this is for
        columnar data
    title : Union[str, None], optional
        Title of the figure, by default None

    Returns
    -------
    go.Figure
        A plotly figure

    Raises
    ------
    ValueError
        If the ylabels specification does not match the subplot specification
    """
    from plotly.subplots import make_subplots

    if set(subplots) != set(y_labels.keys()):
        raise ValueError(f"Mismatch between ylabels {y_labels} and subplots {subplots}")
    if title is None:
        title = "Data plot"

    n_subplots = len(subplots)
    fig = make_subplots(
        rows=n_subplots,
        cols=1,
        shared_xaxes=True,
        vertical_spacing=0.04,
        subplot_titles=subplots,
    )
    fig.update_layout(title_text=title)
    fig.update_xaxes(title_text=x_label, row=n_subplots, col=1)
    for idx, subplot in enumerate(subplots):
        fig.update_yaxes(title_text=y_labels[subplot], row=idx + 1, col=1)
    return fig


def plot_columns_1d(
    fig,
    data: ResisticsData,
    subplots: List[str],
    subplot_columns: Dict[str, List[str]],
    max_pts: Union[int, None] = 5000,
    label_prefix: str = "",
) -> None:
    """View timeseries data as a line plot"""
    if label_prefix != "":
        label_prefix = f"{label_prefix} : "
    for idx, subplot in enumerate(subplots):
        for column in subplot_columns[subplot]:
            if max_pts is not None:
                nx, ny = lttb_downsample(
                    np.arange(data.x_size()), data[column], max_pts
                )
                nx = data.get_x(samples=nx)
            else:
                nx = data.get_x()
                ny = data[column]
            lineplot = go.Scattergl(x=nx, y=ny, name=f"{label_prefix}{column}")
            fig.add_trace(lineplot, row=idx + 1, col=1)
    return fig



"""
Module containing the resistics configuration

The configuration is an essential part of a resistics environment. It defines
many dependencies, such as which data readers to use for time series data or
calibration data and processing options.

Configuration allows users to insert their own dependencies and processors to
work with data.

Configurations can be saved to and loaded from JSON files.
"""
from typing import List

from resistics.common import ResisticsModel
from resistics.time import TimeReader, TimeReaderNumpy, TimeReaderAscii
from resistics.time import TimeProcess, InterpolateNans, RemoveMean
from resistics.calibrate import Calibrator, SensorCalibrationJSON, SensorCalibrator
from resistics.decimate import DecimationSetup
from resistics.decimate import Decimator
from resistics.window import WindowSetup, Windower
from resistics.spectra import FourierTransform, EvaluationFreqs, SpectraProcess
from resistics.transfunc import TransferFunction, ImpedanceTensor
from resistics.regression import RegressionPreparerGathered
from resistics.regression import Solver, SolverScikitTheilSen


class Configuration(ResisticsModel):
    """
    The resistics configuration

    Configuration can be customised by users who wish to use their own custom
    processes for certain steps. In most cases, customisation will be for:

    - Implementing new time data readers
    - Implementing readers for specific calibration formats
    - Implementing time data processors
    - Implementing spectra data processors
    - Adding new features to extract from the data

    Examples
    --------
    Frequently, configuration will be used to change data readers.

    >>> from resistics.letsgo import get_default_configuration
    >>> config = get_default_configuration()
    >>> config.name
    'default'
    >>> for tr in config.time_readers:
    ...     tr.summary()
    {
        'name': 'TimeReaderAscii',
        'apply_scalings': True,
        'extension': '.txt',
        'delimiter': ' ',
        'n_header': 0
    }
    {
        'name': 'TimeReaderNumpy',
        'apply_scalings': True,
        'extension': '.npy'
    }
    >>> config.sensor_calibrator.summary()
    {
        'name': 'SensorCalibrator',
        'chans': None,
        'readers': [
            {
                'name': 'SensorCalibrationJSON',
                'extension': '.json',
                'file_str': 'IC_$sensor$extension'
            }
        ]
    }

    To change these, it's best to make a new configuration with a different name

    >>> from resistics.letsgo import Configuration
    >>> from resistics.time import TimeReaderNumpy
    >>> config = Configuration(name="myconfig", time_readers=[TimeReaderNumpy(apply_scalings=False)])
    >>> for tr in config.time_readers:
    ...     tr.summary()
    {
        'name': 'TimeReaderNumpy',
        'apply_scalings': False,
        'extension': '.npy'
    }

    Or for the sensor calibration

    >>> from resistics.calibrate import SensorCalibrator, SensorCalibrationTXT
    >>> calibration_reader = SensorCalibrationTXT(file_str="lemi120_IC_$serial$extension")
    >>> calibrator = SensorCalibrator(chans=["Hx", "Hy", "Hz"], readers=[calibration_reader])
    >>> config = Configuration(name="myconfig", sensor_calibrator=calibrator)
    >>> config.sensor_calibrator.summary()
    {
        'name': 'SensorCalibrator',
        'chans': ['Hx', 'Hy', 'Hz'],
        'readers': [
            {
                'name': 'SensorCalibrationTXT',
                'extension': '.TXT',
                'file_str': 'lemi120_IC_$serial$extension'
            }
        ]
    }

    As a final example, create a configuration which used targetted windowing
    instead of specified window sizes

    >>> from resistics.letsgo import Configuration
    >>> from resistics.window import WindowerTarget
    >>> config = Configuration(name="window_target", windower=WindowerTarget(target=500))
    >>> config.name
    'window_target'
    >>> config.windower.summary()
    {
        'name': 'WindowerTarget',
        'target': 500,
        'min_size': 64,
        'olap_proportion': 0.25
    }
    """

    name: str
    """The name of the configuration"""
    time_readers: List[TimeReader] = [TimeReaderAscii(), TimeReaderNumpy()]
    """Time readers in the configuration"""
    time_processors: List[TimeProcess] = [InterpolateNans(), RemoveMean()]
    """List of time processors to run"""
    dec_setup: DecimationSetup = DecimationSetup()
    """Process to calculate decimation parameters"""
    decimator: Decimator = Decimator()
    """Process to decimate time data"""
    win_setup: WindowSetup = WindowSetup()
    """Process to calculate windowing parameters"""
    windower: Windower = Windower()
    """Process to window the decimated data"""
    fourier: FourierTransform = FourierTransform()
    """Process to perform the fourier transform"""
    spectra_processors: List[SpectraProcess] = []
    """List of processors to run on spectra data"""
    evals: EvaluationFreqs = EvaluationFreqs()
    """Process to get the spectra data at the evaluation frequencies"""
    sensor_calibrator: Calibrator = SensorCalibrator(readers=[SensorCalibrationJSON()])
    """The sensor calibrator and associated calibration file readers"""
    tf: TransferFunction = ImpedanceTensor()
    """The transfer function to solve"""
    regression_preparer: RegressionPreparerGathered = RegressionPreparerGathered()
    """Process to prepare linear equations"""
    solver: Solver = SolverScikitTheilSen()
    """The solver to use to estimate the regression parameters"""


def get_default_configuration() -> Configuration:
    """Get the default configuration"""
    return Configuration(name="default")
